'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
import math

from pm4py.algo.conformance.alignments.process_tree.variants.approximated.calculate_a_sa_ea_sets import (
    initialize_a_sa_ea_tau_sets, )
from pm4py.algo.conformance.alignments.process_tree.variants.approximated.utilities import (
    calculate_optimal_alignment,
    concatenate_traces,
    trace_to_list_of_str,
    add_fitness_and_cost_info_to_alignments,
    AlignmentNoneException,
    EfficientTree,
)
from pm4py.objects.process_tree.utils.generic import (
    get_process_tree_height,
    process_tree_to_binary_process_tree,
)
from pm4py.objects.petri_net.utils.align_utils import SKIP
from pm4py.objects.process_tree.obj import ProcessTree
from pm4py.objects.log.obj import Trace
from pm4py.objects.log.obj import EventLog
from typing import Union, Dict, Set, List, Tuple
from pm4py.objects.process_tree.obj import Operator
from pm4py.util.xes_constants import DEFAULT_NAME_KEY
from pm4py.util import exec_utils, constants
from pm4py.statistics.variants.log.get import get_variants_from_log_trace_idx
from enum import Enum
import sys
import time
from pm4py.util import variants_util


class Parameters(Enum):
    MAX_TRACE_LENGTH = "max_trace_length"
    MAX_PROCESS_TREE_HEIGHT = "max_process_tree_height"
    PARAMETER_VARIANT_DELIMITER = "variant_delimiter"
    ACTIVITY_KEY = constants.PARAMETER_CONSTANT_ACTIVITY_KEY
    PARAM_MAX_ALIGN_TIME_TRACE = "max_align_time_trace"
    PARAM_MAX_ALIGN_TIME = "max_align_time"


def apply_from_variants_tree_string(var_list, tree_string, parameters=None):
    """
    Apply the alignments from the specification of a list of variants in the log.
    The tree is specified as a PTML input

    Parameters
    ------------
    var_list
        List of variants (for each item, the first entry is the variant itself, the second entry may be the number of cases)
    tree_string
        PTML string representing the tree
    parameters
        Parameters of the algorithm

        Returns
    --------------
    dictio_alignments
        Dictionary that assigns to each variant its alignment
    """
    if parameters is None:
        parameters = {}

    from pm4py.objects.process_tree.importer.variants import ptml

    tree = ptml.import_tree_from_string(tree_string, parameters=parameters)

    res = apply_from_variants_list(var_list, tree, parameters=parameters)
    return res


def apply_from_variants_list(var_list, tree, parameters=None):
    """
    Apply the alignments from the specification of a list of variants in the log

    Parameters
    -------------
    var_list
        List of variants (for each item, the first entry is the variant itself, the second entry may be the number of cases)
    tree
        Process tree
    parameters
        Parameters of the algorithm

    Returns
    --------------
    dictio_alignments
        Dictionary that assigns to each variant its alignment
    """
    if parameters is None:
        parameters = {}

    dictio_alignments = {}
    log = EventLog()

    for index, varitem in enumerate(var_list):
        trace = variants_util.variant_to_trace(
            varitem[0], parameters=parameters
        )
        log.append(trace)

    alignments = apply(log, tree, parameters=parameters)
    for index, varitem in enumerate(var_list):
        dictio_alignments[varitem[0]] = alignments[index]
    return dictio_alignments


def apply(obj: Union[Trace, EventLog], pt: ProcessTree, parameters=None):
    """
    Returns approximated alignments for a process tree

    Parameters
    --------------
    obj
        Event log or trace (a conversion is done if necessary)
    pt
        Process tree
    parameters
        Parameters of the algorithm

    Returns
    --------------
    alignments
        Approximated alignments
    :param obj:
    :param pt:
    :param parameters:
    :return:
    """
    if parameters is None:
        parameters = {}

    max_trace_length = exec_utils.get_param_value(
        Parameters.MAX_TRACE_LENGTH, parameters, 1
    )
    max_process_tree_height = exec_utils.get_param_value(
        Parameters.MAX_PROCESS_TREE_HEIGHT, parameters, 1
    )

    return __align(
        obj,
        pt,
        max_trace_length=max_trace_length,
        max_process_tree_height=max_process_tree_height,
        parameters=parameters,
    )


def __align(
    obj: Union[Trace, EventLog],
    pt: ProcessTree,
    max_trace_length: int = 1,
    max_process_tree_height: int = 1,
    parameters=None,
):
    """
    this function approximates alignments for a given event log or trace and a process tree

    :param obj: event log or single trace
    :param pt: process tree
    :param max_trace_length: specifies when the recursive splitting stops based on the trace's length
    :param max_process_tree_height: specifies when the recursive splitting stops based on the tree's height
    :return:
    """
    assert isinstance(pt, ProcessTree)
    if isinstance(obj, Trace):
        e = EventLog()
        e.append(obj)
        obj = e
    assert isinstance(obj, EventLog)
    pt = process_tree_to_binary_process_tree(pt)
    pt = EfficientTree(pt)

    return __approximate_alignments_for_log(
        obj,
        pt,
        max_trace_length,
        max_process_tree_height,
        parameters=parameters,
    )


def __approximate_alignments_for_log(
    log: EventLog, pt: ProcessTree, max_tl: int, max_th: int, parameters=None
):
    if parameters is None:
        parameters = {}

    a_sets, sa_sets, ea_sets, tau_sets = initialize_a_sa_ea_tau_sets(pt)
    variants = get_variants_from_log_trace_idx(log, parameters=parameters)
    inv_corr = {}

    max_align_time = exec_utils.get_param_value(
        Parameters.PARAM_MAX_ALIGN_TIME, parameters, sys.maxsize
    )
    log_alignment_start_time = time.time()

    for i, var in enumerate(variants):
        this_time = time.time()

        if this_time - log_alignment_start_time <= max_align_time:
            parameters["trace_alignment_start_time"] = this_time
            alignment = __approximate_alignment_for_trace(
                pt,
                a_sets,
                sa_sets,
                ea_sets,
                tau_sets,
                log[variants[var][0]],
                max_tl,
                max_th,
                parameters=parameters,
            )
            alignment = add_fitness_and_cost_info_to_alignments(
                alignment, pt, log[variants[var][0]], parameters=parameters
            )
        else:
            alignment = None

        for idx in variants[var]:
            inv_corr[idx] = alignment
    alignments = []
    for i in range(len(log)):
        alignments.append(inv_corr[i])
    return alignments


def __approximate_alignment_for_trace(
    pt: ProcessTree,
    a_sets: Dict[ProcessTree, Set[str]],
    sa_sets: Dict[ProcessTree, Set[str]],
    ea_sets: Dict[ProcessTree, Set[str]],
    tau_flags: Dict[ProcessTree, bool],
    trace: Trace,
    max_tl: int,
    max_th: int,
    parameters=None,
):
    if parameters is None:
        parameters = {}

    max_align_time_trace = exec_utils.get_param_value(
        Parameters.PARAM_MAX_ALIGN_TIME_TRACE, parameters, sys.maxsize
    )

    start_time = parameters["trace_alignment_start_time"]
    current_time = time.time()

    if current_time - start_time > max_align_time_trace:
        # the alignment of the trace did not terminate in an useful time
        return None

    try:
        if len(trace) <= max_tl or get_process_tree_height(pt) <= max_th:
            return calculate_optimal_alignment(
                pt, trace, parameters=parameters
            )
        else:
            if pt.operator == Operator.SEQUENCE:
                return __approximate_alignment_on_sequence(
                    pt,
                    trace,
                    a_sets,
                    sa_sets,
                    ea_sets,
                    tau_flags,
                    max_tl,
                    max_th,
                    parameters=parameters,
                )
            elif pt.operator == Operator.LOOP:
                return __approximate_alignment_on_loop(
                    pt,
                    trace,
                    a_sets,
                    sa_sets,
                    ea_sets,
                    tau_flags,
                    max_tl,
                    max_th,
                    parameters=parameters,
                )
            elif pt.operator == Operator.XOR:
                return __approximate_alignment_on_choice(
                    pt,
                    trace,
                    a_sets,
                    sa_sets,
                    ea_sets,
                    tau_flags,
                    max_tl,
                    max_th,
                    parameters=parameters,
                )
            elif pt.operator == Operator.PARALLEL:
                return __approximate_alignment_on_parallel(
                    pt,
                    trace,
                    a_sets,
                    sa_sets,
                    ea_sets,
                    tau_flags,
                    max_tl,
                    max_th,
                    parameters=parameters,
                )
    except AlignmentNoneException:
        # alignment did not terminate correctly. return None
        return None


def __approximate_alignment_on_choice(
    pt: ProcessTree,
    trace: Trace,
    a_sets: Dict[ProcessTree, Set[str]],
    sa_sets: Dict[ProcessTree, Set[str]],
    ea_sets: Dict[ProcessTree, Set[str]],
    tau_flags: Dict[ProcessTree, bool],
    tl: int,
    th: int,
    parameters=None,
):
    if parameters is None:
        parameters = {}
    activity_key = exec_utils.get_param_value(
        Parameters.ACTIVITY_KEY, parameters, DEFAULT_NAME_KEY
    )

    assert pt.operator == Operator.XOR
    assert len(trace) > 0

    best_suited_subtree = None
    lowest_mismatches = math.inf
    for subtree in pt.children:
        mismatches = 0
        if len(trace) > 0:
            if trace[0][activity_key] not in sa_sets[subtree]:
                mismatches += 1
            if trace[-1][activity_key] not in ea_sets[subtree]:
                mismatches += 1
            if len(trace) > 2:
                for a in trace[1:-1]:
                    if a[activity_key] not in a_sets[subtree]:
                        mismatches += 1
        else:
            if (
                not tau_flags[subtree]
                and len(sa_sets[subtree].intersection(ea_sets[subtree])) != 0
            ):
                mismatches += 1
            elif (
                not tau_flags[subtree]
                and len(sa_sets[subtree].intersection(ea_sets[subtree])) == 0
            ):
                mismatches += 2
        if mismatches < lowest_mismatches:
            best_suited_subtree = subtree
            lowest_mismatches = mismatches
    return __approximate_alignment_for_trace(
        best_suited_subtree,
        a_sets,
        sa_sets,
        ea_sets,
        tau_flags,
        trace,
        tl,
        th,
        parameters=parameters,
    )


def __approximate_alignment_on_loop(
    pt: ProcessTree,
    trace: Trace,
    a_sets: Dict[ProcessTree, Set[str]],
    sa_sets: Dict[ProcessTree, Set[str]],
    ea_sets: Dict[ProcessTree, Set[str]],
    tau_flags: Dict[ProcessTree, bool],
    tl: int,
    th: int,
    parameters=None,
):
    if parameters is None:
        parameters = {}

    from pulp import lpSum, LpVariable, LpProblem, LpMinimize

    activity_key = exec_utils.get_param_value(
        Parameters.ACTIVITY_KEY, parameters, DEFAULT_NAME_KEY
    )

    assert pt.operator == Operator.LOOP
    assert len(pt.children) == 2
    assert len(trace) > 0

    ilp = LpProblem(sense=LpMinimize)

    # x_i_j = 1 <=> assigns activity i to subtree j
    x_variables = {}

    # t_i_j = 1 <=> inserts a tau at position i and assigns it to subtree j
    t_variables = {}

    # s_i_j = 1 <=> activity i is a start activity in the current sub-trace
    # assigned to subtree j
    s_variables = {}

    # e_i_j = 1 <=> activity i is an end activity in the current sub-trace
    # assigned to subtree j
    e_variables = {}

    # v_i_j = 1 <=> activity i is neither a start nor end-activity in the
    # current sub-trace assigned to subtree j
    v_variables = {}

    # auxiliary variables
    # p_i_j = 1 <=> previous activity i-1 is assigned to the other subtree or
    # t_1_other-subtree is 1
    p_variables = {}

    # n_i_j = 1 <=> next activity i+1 is assigned to the other subtree or
    # t_1_other-subtree is 1
    n_variables = {}

    t_costs = {}
    s_costs = {}
    e_costs = {}
    v_costs = {}

    for i, a in enumerate(trace):
        x_variables[i] = {}
        s_variables[i] = {}
        s_costs[i] = {}
        e_variables[i] = {}
        e_costs[i] = {}
        v_variables[i] = {}
        v_costs[i] = {}
        p_variables[i] = {}
        n_variables[i] = {}
        for j, subtree in enumerate(pt.children):
            x_variables[i][j] = LpVariable(
                "x_" + str(i) + "_" + str(j), cat="Binary"
            )

            s_variables[i][j] = LpVariable(
                "s_" + str(i) + "_" + str(j), cat="Binary"
            )
            s_costs[i][j] = 0 if a[activity_key] in sa_sets[subtree] else 1

            e_variables[i][j] = LpVariable(
                "e_" + str(i) + "_" + str(j), cat="Binary"
            )
            e_costs[i][j] = 0 if a[activity_key] in ea_sets[subtree] else 1

            v_variables[i][j] = LpVariable(
                "v_" + str(i) + "_" + str(j), cat="Binary"
            )
            v_costs[i][j] = 0 if a[activity_key] in a_sets[subtree] else 1

            p_variables[i][j] = LpVariable(
                "p_" + str(i) + "_" + str(j), cat="Binary"
            )
            n_variables[i][j] = LpVariable(
                "n_" + str(i) + "_" + str(j), cat="Binary"
            )

    for i in range(len(trace) + 1):
        t_variables[i] = {}
        t_costs[i] = {}
        for j, subtree in enumerate(pt.children):
            t_variables[i][j] = LpVariable(
                "t_" + str(i) + "_" + str(j), cat="Binary"
            )
            if tau_flags[subtree]:
                t_costs[i][
                    j
                ] = (
                    -0.00001
                )  # favour to add a cut if possible over not putting a cut
            else:
                if len(sa_sets[subtree].intersection(ea_sets[subtree])) != 0:
                    t_costs[i][j] = 1
                else:
                    t_costs[i][j] = 2

    # objective function
    ilp += (
        lpSum(
            [
                s_variables[i][j] * s_costs[i][j]
                for i in range(len(trace))
                for j in range(len(pt.children))
            ]
            + [
                e_variables[i][j] * e_costs[i][j]
                for i in range(len(trace))
                for j in range(len(pt.children))
            ]
            + [
                v_variables[i][j] * v_costs[i][j]
                for i in range(len(trace))
                for j in range(len(pt.children))
            ]
            + [
                t_variables[i][j] * t_costs[i][j]
                for i in range(len(trace) + 1)
                for j in range(len(pt.children))
            ]
        ),
        "objective_function",
    )

    # constraints
    # universe j                        {0,1}
    # universe i for t_i_j variables    {0,...,len(trace)}
    # universe i else                   {0,...,len(trace)-1}

    # first tau can never be assigned to the 2nd subtree
    ilp += t_variables[0][1] == 0

    # last tau can never be assigned to the 2nd subtree
    ilp += t_variables[len(trace)][1] == 0

    # if first/last tau is not used --> first/last activity is assigned to 1st
    # subtree
    ilp += 1 - t_variables[0][0] <= x_variables[0][0]
    ilp += 1 - t_variables[len(trace)][0] <= x_variables[len(trace) - 1][0]

    for i in range(len(trace)):
        # every activity is assigned to one subtree
        ilp += (
            lpSum([x_variables[i][j] * 1 for j in range(len(pt.children))])
            == 1
        )

        # start/end/intermediate-activity at position i can only be assigned to
        # one subtree
        ilp += (
            lpSum([s_variables[i][j] * 1 for j in range(len(pt.children))])
            <= 1
        )
        ilp += (
            lpSum([e_variables[i][j] * 1 for j in range(len(pt.children))])
            <= 1
        )
        ilp += (
            lpSum([v_variables[i][j] * 1 for j in range(len(pt.children))])
            <= 1
        )

    for i in range(len(trace) + 1):
        # max one tau is used per index
        ilp += lpSum([t_variables[i][j] for j in range(2)]) <= 1

    # if tau is used and hence, assigned to a subtree, the surrounding
    # activities are assigned to the other subtree
    for i in range(1, len(trace)):
        # if tau at position i is assigned to 1st subtree, the previous
        # activity is assigned to 2nd subtree
        ilp += t_variables[i][0] <= x_variables[i - 1][1]
        # if tau at position i is assigned to 1st subtree, the previous
        # activity is assigned to 2nd subtree
        ilp += t_variables[i][1] <= x_variables[i - 1][0]
    for i in range(len(trace)):
        # if tau at position i is assigned to 1st subtree, the next activity is
        # assigned to 2nd subtree
        ilp += t_variables[i][0] <= x_variables[i][1]
        # if tau at position i is assigned to 2nd subtree, the next activity is
        # assigned to 1st subtree
        ilp += t_variables[i][1] <= x_variables[i][0]
    # if last tau is used and assigned to 1st subtree (assigning it to the 2nd subtree is already forbidden by another
    # constraint) --> last activity must be assigned to 2nd subtree
    ilp += t_variables[len(trace)][0] <= x_variables[len(trace) - 1][1]

    # define auxiliary variables n: n_i_1 = 1 <=> next activity i+1 is
    # assigned to 2nd subtree or t_i+1_2 = 1
    for i in range(len(trace) - 1):
        ilp += (
            n_variables[i][0] <= x_variables[i + 1][1] + t_variables[i + 1][1]
        )
        ilp += n_variables[i][0] >= x_variables[i + 1][1]
        ilp += n_variables[i][0] >= t_variables[i + 1][1]

        ilp += (
            n_variables[i][1] <= x_variables[i + 1][0] + t_variables[i + 1][0]
        )
        ilp += n_variables[i][1] >= x_variables[i + 1][0]
        ilp += n_variables[i][1] >= t_variables[i + 1][0]

    ilp += t_variables[len(trace)][1] <= n_variables[len(trace) - 1][0]
    ilp += t_variables[len(trace)][0] <= n_variables[len(trace) - 1][1]

    # define e_i_j variables
    for i in range(len(trace)):
        for j in range(2):
            ilp += e_variables[i][j] <= n_variables[i][j]
            ilp += e_variables[i][j] <= x_variables[i][j]
            ilp += (
                e_variables[i][j] >= n_variables[i][j] + x_variables[i][j] - 1
            )

    # define auxiliary variables p: p_i_1 = 1 <=> previous activity i-1 is
    # assigned to 2nd subtree or t_i-1_2 = 1
    ilp += t_variables[0][1] <= p_variables[0][0]
    ilp += p_variables[0][1] <= t_variables[0][0]

    for i in range(1, len(trace)):
        ilp += p_variables[i][0] <= t_variables[i][1] + x_variables[i - 1][1]
        ilp += p_variables[i][0] >= t_variables[i][1]
        ilp += p_variables[i][0] >= x_variables[i - 1][1]

        ilp += p_variables[i][1] <= t_variables[i][0] + x_variables[i - 1][0]
        ilp += p_variables[i][1] >= t_variables[i][0]
        ilp += p_variables[i][1] >= x_variables[i - 1][0]

    # define s_i_j variables
    for i in range(len(trace)):
        for j in range(2):
            ilp += (
                s_variables[i][j] >= p_variables[i][j] + x_variables[i][j] - 1
            )
            ilp += s_variables[i][j] <= p_variables[i][j]
            ilp += s_variables[i][j] <= p_variables[i][j]
    ilp += 1 - t_variables[0][0] <= s_variables[0][0]

    # define v_i_j variables
    for i in range(len(trace)):
        for j in range(2):
            ilp += (
                v_variables[i][j]
                >= 1
                - s_variables[i][j]
                + 1
                - e_variables[i][j]
                + x_variables[i][j]
                - 2
            )
            ilp += v_variables[i][j] <= x_variables[i][j]
            ilp += v_variables[i][j] <= 1 - e_variables[i][j]
            ilp += v_variables[i][j] <= 1 - s_variables[i][j]

    status = ilp.solve()
    assert status == 1

    alignments_to_calculate = []
    sub_trace = Trace()
    current_subtree_idx = 0
    for i in range(len(trace)):
        for j in range(2):
            if t_variables[i][j].varValue:
                if i == 0:
                    # first tau can be only assigned to first subtree
                    assert j == 0
                    alignments_to_calculate.append((pt.children[j], Trace()))
                    current_subtree_idx = 1
                else:
                    alignments_to_calculate.append(
                        (pt.children[current_subtree_idx], sub_trace)
                    )
                    alignments_to_calculate.append((pt.children[j], Trace()))
                    sub_trace = Trace()
        for j in range(2):
            if x_variables[i][j].varValue:
                if j == current_subtree_idx:
                    sub_trace.append(trace[i])
                else:
                    alignments_to_calculate.append(
                        (pt.children[current_subtree_idx], sub_trace)
                    )
                    sub_trace = Trace()
                    sub_trace.append(trace[i])
                    current_subtree_idx = j
    if len(sub_trace) > 0:
        alignments_to_calculate.append(
            (pt.children[current_subtree_idx], sub_trace)
        )
    if t_variables[len(trace)][0].varValue:
        alignments_to_calculate.append((pt.children[0], Trace()))

    res = []
    for subtree, sub_trace in alignments_to_calculate:
        align_result = __approximate_alignment_for_trace(
            subtree,
            a_sets,
            sa_sets,
            ea_sets,
            tau_flags,
            sub_trace,
            tl,
            th,
            parameters=parameters,
        )
        if align_result is None:
            # the alignment did not terminate correctly.
            return None
        res.extend(align_result)
    return res


def __approximate_alignment_on_sequence(
    pt: ProcessTree,
    trace: Trace,
    a_sets: Dict[ProcessTree, Set[str]],
    sa_sets: Dict[ProcessTree, Set[str]],
    ea_sets: Dict[ProcessTree, Set[str]],
    tau_flags: Dict[ProcessTree, bool],
    tl: int,
    th: int,
    parameters=None,
):
    if parameters is None:
        parameters = {}

    from pulp import lpSum, LpVariable, LpProblem, LpMinimize

    activity_key = exec_utils.get_param_value(
        Parameters.ACTIVITY_KEY, parameters, DEFAULT_NAME_KEY
    )

    assert pt.operator == Operator.SEQUENCE
    assert len(pt.children) > 0
    assert len(trace) > 0

    ilp = LpProblem(sense=LpMinimize)

    # x_i_j = 1 <=> assigns activity i to subtree j
    x_variables = {}

    # s_i_j = 1 <=> activity i is a start activity in the current sub-trace
    # assigned to subtree j
    s_variables = {}

    # e_i_j = 1 <=> activity i is an end activity in the current sub-trace
    # assigned to subtree j
    e_variables = {}

    # auxiliary u_j <=> u_j=1 if an activity is assigned to subtree j
    u_variables = {}

    # v_i_j = 1 <=> activity i is neither a start nor end-activity in the
    # current sub-trace assigned to subtree j
    v_variables = {}

    s_costs = {}
    e_costs = {}
    u_costs = {}
    v_costs = {}

    # trace <a_0,...,a_n>
    for i, a in enumerate(trace):
        x_variables[i] = {}
        s_variables[i] = {}
        s_costs[i] = {}
        e_variables[i] = {}
        e_costs[i] = {}
        v_variables[i] = {}
        v_costs[i] = {}

        for j, subtree in enumerate(pt.children):
            x_variables[i][j] = LpVariable(
                "x_" + str(i) + "_" + str(j), cat="Binary"
            )

            s_variables[i][j] = LpVariable(
                "s_" + str(i) + "_" + str(j), cat="Binary"
            )
            s_costs[i][j] = 0 if a[activity_key] in sa_sets[subtree] else 1

            e_variables[i][j] = LpVariable(
                "e_" + str(i) + "_" + str(j), cat="Binary"
            )
            e_costs[i][j] = 0 if a[activity_key] in ea_sets[subtree] else 1

            v_variables[i][j] = LpVariable(
                "v_" + str(i) + "_" + str(j), cat="Binary"
            )
            v_costs[i][j] = 0 if a[activity_key] in a_sets[subtree] else 1

    for j in range(len(pt.children)):
        u_variables[j] = LpVariable("u_" + str(j), cat="Binary")
        # define costs to not assign anything to subtree j
        if tau_flags[pt.children[j]]:
            u_costs[j] = 0
        elif sa_sets[pt.children[j]] & ea_sets[pt.children[j]]:
            # intersection of start-activities and end-activities is not empty
            u_costs[j] = 1
        else:
            # intersection of start-activities and end-activities is empty
            u_costs[j] = 2

    # objective function
    ilp += (
        lpSum(
            [
                v_variables[i][j] * v_costs[i][j]
                for i in range(len(trace))
                for j in range(len(pt.children))
            ]
            + [
                s_variables[i][j] * s_costs[i][j]
                for i in range(len(trace))
                for j in range(len(pt.children))
            ]
            + [
                e_variables[i][j] * e_costs[i][j]
                for i in range(len(trace))
                for j in range(len(pt.children))
            ]
            + [
                (1 - u_variables[j]) * u_costs[j]
                for j in range(len(pt.children))
            ]
        ),
        "objective_function",
    )

    # constraints
    for i in range(len(trace)):
        # every activity is assigned to one subtree
        ilp += (
            lpSum([x_variables[i][j] * 1 for j in range(len(pt.children))])
            == 1
        )

    for j in range(len(pt.children)):
        # first activity is start activity
        ilp += x_variables[0][j] <= s_variables[0][j]
        # last activity is end-activity
        ilp += x_variables[len(trace) - 1][j] <= e_variables[len(trace) - 1][j]

    # define s_i_j variables
    for i in range(1, len(trace)):
        for j in range(len(pt.children)):
            ilp += (
                s_variables[i][j]
                >= x_variables[i][j] + 1 - x_variables[i - 1][j] - 1
            )
            ilp += s_variables[i][j] <= x_variables[i][j]
            ilp += s_variables[i][j] <= 1 - x_variables[i - 1][j]
    for i in range(len(trace)):
        # activity can be only for one subtree a start-activity
        ilp += lpSum(s_variables[i][j] for j in range(len(pt.children))) <= 1

    # define e_i_j variables
    for i in range(len(trace) - 1):
        for j in range(len(pt.children)):
            ilp += (
                e_variables[i][j]
                >= x_variables[i][j] + 1 - x_variables[i + 1][j] - 1
            )
            ilp += e_variables[i][j] <= x_variables[i][j]
            ilp += e_variables[i][j] <= 1 - x_variables[i + 1][j]
    for i in range(len(trace)):
        # activity can be only for one subtree an end-activity
        ilp += lpSum(e_variables[i][j] for j in range(len(pt.children))) <= 1

    # constraint - preserving sequence when assigning activities to subtrees
    for i in range(len(trace) - 1):
        for j in range(len(pt.children)):
            ilp += (
                lpSum(
                    x_variables[i + 1][k] for k in range(j, len(pt.children))
                )
                >= x_variables[i][j]
            )

    for j in range(len(pt.children)):
        for i in range(len(trace)):
            # define u_j variables
            ilp += u_variables[j] >= x_variables[i][j]

        # if u_j variable = 1 ==> a start activity must exist
        ilp += u_variables[j] <= lpSum(
            s_variables[i][j] for i in range(len(trace))
        )
        # if u_j variable = 1 ==> an end activity must exist
        ilp += u_variables[j] <= lpSum(
            e_variables[i][j] for i in range(len(trace))
        )

    # define v_i_j variables
    for i in range(len(trace)):
        for j in range(2):
            ilp += (
                v_variables[i][j]
                >= 1
                - s_variables[i][j]
                + 1
                - e_variables[i][j]
                + x_variables[i][j]
                - 2
            )
            ilp += v_variables[i][j] <= x_variables[i][j]
            ilp += v_variables[i][j] <= 1 - e_variables[i][j]
            ilp += v_variables[i][j] <= 1 - s_variables[i][j]

    status = ilp.solve()
    assert status == 1

    alignments_to_calculate = []
    for j in range(len(pt.children)):
        sub_trace = Trace()
        for i in range(len(trace)):
            if x_variables[i][j].varValue == 1:
                sub_trace.append(trace[i])
        alignments_to_calculate.append((pt.children[j], sub_trace))
    # calculate and compose alignments
    res = []
    for subtree, sub_trace in alignments_to_calculate:
        align_result = __approximate_alignment_for_trace(
            subtree,
            a_sets,
            sa_sets,
            ea_sets,
            tau_flags,
            sub_trace,
            tl,
            th,
            parameters=parameters,
        )
        if align_result is None:
            # the alignment did not terminate correctly.
            return None
        res.extend(align_result)
    return res


def __approximate_alignment_on_parallel(
    pt: ProcessTree,
    trace: Trace,
    a_sets: Dict[ProcessTree, Set[str]],
    sa_sets: Dict[ProcessTree, Set[str]],
    ea_sets: Dict[ProcessTree, Set[str]],
    tau_flags: Dict[ProcessTree, bool],
    tl: int,
    th: int,
    parameters=None,
):
    if parameters is None:
        parameters = {}

    from pulp import lpSum, LpVariable, LpProblem, LpMinimize

    activity_key = exec_utils.get_param_value(
        Parameters.ACTIVITY_KEY, parameters, DEFAULT_NAME_KEY
    )

    assert pt.operator == Operator.PARALLEL
    assert len(pt.children) > 0
    assert len(trace) > 0

    ilp = LpProblem(sense=LpMinimize)

    # x_i_j = 1 <=> assigns activity i to subtree j
    x_variables = {}

    # s_i_j = 1 <=> activity i is a start activity in the current sub-trace
    # assigned to subtree j
    s_variables = {}

    # e_i_j = 1 <=> activity i is an end activity in the current sub-trace
    # assigned to subtree j
    e_variables = {}

    # auxiliary u_j <=> u_j=1 if an activity is assigned to subtree j
    u_variables = {}

    # v_i_j = 1 <=> activity i is neither a start nor end-activity in the
    # current sub-trace assigned to subtree j
    v_variables = {}

    s_costs = {}
    e_costs = {}
    u_costs = {}
    v_costs = {}

    for i, a in enumerate(trace):
        x_variables[i] = {}
        s_variables[i] = {}
        s_costs[i] = {}
        e_variables[i] = {}
        e_costs[i] = {}
        v_variables[i] = {}
        v_costs[i] = {}

        for j, subtree in enumerate(pt.children):
            x_variables[i][j] = LpVariable(
                "x_" + str(i) + "_" + str(j), cat="Binary"
            )

            s_variables[i][j] = LpVariable(
                "s_" + str(i) + "_" + str(j), cat="Binary"
            )
            s_costs[i][j] = 0 if a[activity_key] in sa_sets[subtree] else 1

            e_variables[i][j] = LpVariable(
                "e_" + str(i) + "_" + str(j), cat="Binary"
            )
            e_costs[i][j] = 0 if a[activity_key] in ea_sets[subtree] else 1

            v_variables[i][j] = LpVariable(
                "v_" + str(i) + "_" + str(j), cat="Binary"
            )
            v_costs[i][j] = 0 if a[activity_key] in a_sets[subtree] else 1

    for j in range(len(pt.children)):
        u_variables[j] = LpVariable("u_" + str(j), cat="Binary")
        # define costs to not assign anything to subtree j
        if tau_flags[pt.children[j]]:
            u_costs[j] = 0
        elif sa_sets[pt.children[j]] & ea_sets[pt.children[j]]:
            # intersection of start-activities and end-activities is not empty
            u_costs[j] = 1
        else:
            # intersection of start-activities and end-activities is empty
            u_costs[j] = 2

    # objective function
    ilp += (
        lpSum(
            [
                v_variables[i][j] * v_costs[i][j]
                for i in range(len(trace))
                for j in range(len(pt.children))
            ]
            + [
                s_variables[i][j] * s_costs[i][j]
                for i in range(len(trace))
                for j in range(len(pt.children))
            ]
            + [
                e_variables[i][j] * e_costs[i][j]
                for i in range(len(trace))
                for j in range(len(pt.children))
            ]
            + [
                (1 - u_variables[j]) * u_costs[j]
                for j in range(len(pt.children))
            ]
        ),
        "objective_function",
    )

    # constraints
    for i in range(len(trace)):
        # every activity is assigned to one subtree
        ilp += (
            lpSum([x_variables[i][j] * 1 for j in range(len(pt.children))])
            == 1
        )

    for j in range(len(pt.children)):
        # first activity is a start activity
        ilp += x_variables[0][j] <= s_variables[0][j]
        # last activity is an end-activity
        ilp += x_variables[len(trace) - 1][j] <= e_variables[len(trace) - 1][j]

    # define s_i_j variables
    for i in range(len(trace)):
        for j in range(len(pt.children)):
            ilp += s_variables[i][j] <= x_variables[i][j]
            for k in range(i):
                ilp += s_variables[i][j] <= 1 - x_variables[k][j]
        # activity can be only a start-activity for one subtree
        ilp += lpSum(s_variables[i][j] for j in range(len(pt.children))) <= 1

    # define e_i_j variables
    for i in range(len(trace)):
        for j in range(len(pt.children)):
            ilp += e_variables[i][j] <= x_variables[i][j]
            for k in range(i + 1, len(trace)):
                ilp += e_variables[i][j] <= 1 - x_variables[k][j]
        # activity can be only an end-activity for one subtree
        ilp += lpSum(e_variables[i][j] for j in range(len(pt.children))) <= 1

    for j in range(len(pt.children)):
        for i in range(len(trace)):
            # define u_j variables
            ilp += u_variables[j] >= x_variables[i][j]
        # if u_j variable = 1 ==> a start activity must exist
        ilp += u_variables[j] <= lpSum(
            s_variables[i][j] for i in range(len(trace))
        )
        # if u_j variable = 1 ==> an end activity must exist
        ilp += u_variables[j] <= lpSum(
            e_variables[i][j] for i in range(len(trace))
        )

    # define v_i_j variables
    for i in range(len(trace)):
        for j in range(2):
            ilp += (
                v_variables[i][j]
                >= 1
                - s_variables[i][j]
                + 1
                - e_variables[i][j]
                + x_variables[i][j]
                - 2
            )
            ilp += v_variables[i][j] <= x_variables[i][j]
            ilp += v_variables[i][j] <= 1 - e_variables[i][j]
            ilp += v_variables[i][j] <= 1 - s_variables[i][j]

    status = ilp.solve()
    assert status == 1

    # trace_parts list contains trace parts mapped onto the determined subtree
    trace_parts = []
    last_subtree = None
    for i in range(len(trace)):
        for j in range(len(pt.children)):
            subtree = pt.children[j]
            if x_variables[i][j].varValue == 1:
                if last_subtree and subtree == last_subtree:
                    trace_parts[-1][1].append(trace[i])
                else:
                    assert last_subtree is None or subtree != last_subtree
                    t = Trace()
                    t.append(trace[i])
                    trace_parts.append((subtree, t))
                    last_subtree = subtree
                continue

    # calculate an alignment for each subtree
    alignments_per_subtree = {}
    for j in range(len(pt.children)):
        subtree = pt.children[j]
        sub_trace = Trace()
        for trace_part in trace_parts:
            if subtree == trace_part[0]:
                sub_trace = concatenate_traces(sub_trace, trace_part[1])
        align_result = __approximate_alignment_for_trace(
            subtree,
            a_sets,
            sa_sets,
            ea_sets,
            tau_flags,
            sub_trace,
            tl,
            th,
            parameters=parameters,
        )
        if align_result is None:
            # the alignment did not terminate correctly.
            return None
        alignments_per_subtree[subtree] = align_result
    # compose alignments from subtree alignments
    res = []
    for trace_part in trace_parts:
        activities_to_cover = trace_to_list_of_str(trace_part[1])
        activities_covered_so_far = []
        alignment = alignments_per_subtree[trace_part[0]]
        while activities_to_cover != activities_covered_so_far:
            move = alignment.pop(0)
            res.append(move)
            # if the alignment move is NOT a model move add activity to
            # activities_covered_so_far
            if move[0] != SKIP:
                activities_covered_so_far.append(move[0])
    # add possible remaining alignment moves to resulting alignment, the order
    # does not matter (parallel operator)
    for subtree in alignments_per_subtree:
        if len(alignments_per_subtree[subtree]) > 0:
            res.extend(alignments_per_subtree[subtree])
    return res



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from enum import Enum
from pm4py.objects.log.obj import EventLog, EventStream
import pandas as pd
from typing import Optional, Dict, Any, Collection, Union
from pm4py.util import exec_utils, constants, xes_constants, pandas_utils
from pm4py.objects.ocel.obj import OCEL
from pm4py.objects.ocel import constants as ocel_constants
from pm4py.objects.conversion.log import converter as log_converter
from pm4py.objects.ocel.util import ocel_consistency
from copy import copy
import math


class Parameters(Enum):
    CASE_ID_KEY = constants.PARAMETER_CONSTANT_CASEID_KEY
    CASE_ATTRIBUTE_PREFIX = constants.PARAMETER_KEY_CASE_ATTRIBUTE_PRFIX
    ACTIVITY_KEY = constants.PARAMETER_CONSTANT_ACTIVITY_KEY
    TIMESTAMP_KEY = constants.PARAMETER_CONSTANT_TIMESTAMP_KEY
    TARGET_OBJECT_TYPE = "target_object_type"
    TARGET_OBJECT_TYPE_2 = "target_object_type_2"
    LEFT_INDEX = "left_index"
    RIGHT_INDEX = "right_index"
    DIRECTION = "direction"


def __postprocess_stream(list_events):
    """
    Postprocess the list of events of the stream in order to make sure
    that there are no NaN/NaT values

    Parameters
    -------------
    list_events
        List of events

    Returns
    -------------
    list_events
        Postprocessed stream
    """
    for event in list_events:
        event_keys = list(event.keys())
        for k in event_keys:
            typ_k = type(event[k])
            if typ_k is pd._libs.tslibs.nattype.NaTType:
                del event[k]
                continue
            elif (typ_k is float or typ_k is int) and math.isnan(event[k]):
                del event[k]
                continue
            elif event[k] is None:
                del event[k]
                continue
    return list_events


def from_traditional_log(
    log: EventLog, parameters: Optional[Dict[Any, Any]] = None
) -> OCEL:
    """
    Transforms an EventLog to an OCEL

    Parameters
    -----------------
    log
        Event log
    parameters
        Parameters of the algorithm, including:
        - Parameters.TARGET_OBJECT_TYPE => the name of the object type to which the cases should be mapped
        - Parameters.ACTIVITY_KEY => the attribute to use as activity
        - Parameters.TIMESTAMP_KEY => the attribute to use as timestamp
        - Parameters.CASE_ID_KEY => the attribute to use as case identifier

    Returns
    -----------------
    ocel
        OCEL (equivalent to the provided event log)
    """
    if parameters is None:
        parameters = {}

    log = log_converter.apply(
        log, variant=log_converter.Variants.TO_EVENT_LOG, parameters=parameters
    )

    target_object_type = exec_utils.get_param_value(
        Parameters.TARGET_OBJECT_TYPE, parameters, "OTYPE"
    )
    activity_key = exec_utils.get_param_value(
        Parameters.ACTIVITY_KEY, parameters, xes_constants.DEFAULT_NAME_KEY
    )
    timestamp_key = exec_utils.get_param_value(
        Parameters.TIMESTAMP_KEY,
        parameters,
        xes_constants.DEFAULT_TIMESTAMP_KEY,
    )
    case_id_key = exec_utils.get_param_value(
        Parameters.CASE_ID_KEY, parameters, xes_constants.DEFAULT_TRACEID_KEY
    )

    events = []
    objects = []
    relations = []

    ev_count = 0
    for trace in log:
        case_id = trace.attributes[case_id_key]
        obj = {
            ocel_constants.DEFAULT_OBJECT_ID: case_id,
            ocel_constants.DEFAULT_OBJECT_TYPE: target_object_type,
        }
        for attr in trace.attributes:
            if attr != case_id_key:
                obj[attr] = trace.attributes[attr]
        objects.append(obj)
        for ev in trace:
            ev_count = ev_count + 1
            activity = ev[activity_key]
            timestamp = ev[timestamp_key]
            eve = {
                ocel_constants.DEFAULT_EVENT_ID: str(ev_count),
                ocel_constants.DEFAULT_EVENT_ACTIVITY: activity,
                ocel_constants.DEFAULT_EVENT_TIMESTAMP: timestamp,
            }
            for attr in ev:
                if attr not in [activity, timestamp]:
                    eve[attr] = ev[attr]
            events.append(eve)
            relations.append(
                {
                    ocel_constants.DEFAULT_EVENT_ID: str(ev_count),
                    ocel_constants.DEFAULT_EVENT_ACTIVITY: activity,
                    ocel_constants.DEFAULT_EVENT_TIMESTAMP: timestamp,
                    ocel_constants.DEFAULT_OBJECT_ID: case_id,
                    ocel_constants.DEFAULT_OBJECT_TYPE: target_object_type,
                }
            )

    events = pandas_utils.instantiate_dataframe(events)
    objects = pandas_utils.instantiate_dataframe(objects)
    relations = pandas_utils.instantiate_dataframe(relations)

    return OCEL(events=events, objects=objects, relations=relations)


def __get_events_dataframe(
    df: pd.DataFrame,
    activity_key: str,
    timestamp_key: str,
    case_id_key: str,
    case_attribute_prefix: str,
    events_prefix="E",
) -> pd.DataFrame:
    """
    Internal method to get the events dataframe out of a traditional log stored as Pandas dataframe
    """
    columns = {case_id_key}.union(
        set(x for x in df.columns if not x.startswith(case_attribute_prefix))
    )
    columns = list(columns)
    df = df[columns]
    df = df.rename(
        columns={
            activity_key: ocel_constants.DEFAULT_EVENT_ACTIVITY,
            timestamp_key: ocel_constants.DEFAULT_EVENT_TIMESTAMP,
            case_id_key: ocel_constants.DEFAULT_OBJECT_ID,
        }
    )
    df[ocel_constants.DEFAULT_EVENT_ID] = events_prefix + df.index.astype(str)
    return df


def __get_objects_dataframe(
    df: pd.DataFrame,
    case_id_key: str,
    case_attribute_prefix: str,
    target_object_type: str,
) -> pd.DataFrame:
    """
    Internal method to get the objects dataframe out of a traditional log stored as Pandas dataframe
    """
    columns = {x for x in df.columns if x.startswith(case_attribute_prefix)}
    columns = list(columns)
    df = df[columns]
    df = df.rename(columns={case_id_key: ocel_constants.DEFAULT_OBJECT_ID})
    df = df.groupby(ocel_constants.DEFAULT_OBJECT_ID).first().reset_index()
    df[ocel_constants.DEFAULT_OBJECT_TYPE] = target_object_type
    return df


def __get_relations_from_events(
    events: pd.DataFrame, target_object_type: str
) -> pd.DataFrame:
    """
    Internal method to get the relations dataframe out of a traditional log stored as Pandas dataframe
    """
    relations = events[
        [
            ocel_constants.DEFAULT_EVENT_ACTIVITY,
            ocel_constants.DEFAULT_EVENT_TIMESTAMP,
            ocel_constants.DEFAULT_OBJECT_ID,
            ocel_constants.DEFAULT_EVENT_ID,
        ]
    ]
    relations[ocel_constants.DEFAULT_OBJECT_TYPE] = target_object_type
    return relations


def from_traditional_pandas(
    df: pd.DataFrame, parameters: Optional[Dict[Any, Any]] = None
) -> OCEL:
    """
    Transforms a dataframe to an OCEL

    Parameters
    -----------------
    df
        Pandas dataframe
    parameters
        Parameters of the algorithm, including:
        - Parameters.TARGET_OBJECT_TYPE => the name of the object type to which the cases should be mapped
        - Parameters.ACTIVITY_KEY => the attribute to use as activity
        - Parameters.TIMESTAMP_KEY => the attribute to use as timestamp
        - Parameters.CASE_ID_KEY => the attribute to use as case identifier
        - Parameters.CASE_ATTRIBUTE_PREFIX => the prefix identifying the attributes at the case level

    Returns
    -----------------
    ocel
        OCEL (equivalent to the provided event log)
    """
    if parameters is None:
        parameters = {}

    target_object_type = exec_utils.get_param_value(
        Parameters.TARGET_OBJECT_TYPE, parameters, "OTYPE"
    )
    activity_key = exec_utils.get_param_value(
        Parameters.ACTIVITY_KEY, parameters, xes_constants.DEFAULT_NAME_KEY
    )
    timestamp_key = exec_utils.get_param_value(
        Parameters.TIMESTAMP_KEY,
        parameters,
        xes_constants.DEFAULT_TIMESTAMP_KEY,
    )
    case_id_key = exec_utils.get_param_value(
        Parameters.CASE_ID_KEY, parameters, constants.CASE_CONCEPT_NAME
    )
    case_attribute_prefix = exec_utils.get_param_value(
        Parameters.CASE_ATTRIBUTE_PREFIX,
        parameters,
        constants.CASE_ATTRIBUTE_PREFIX,
    )

    events = __get_events_dataframe(
        df, activity_key, timestamp_key, case_id_key, case_attribute_prefix
    )
    objects = __get_objects_dataframe(
        df, case_id_key, case_attribute_prefix, target_object_type
    )
    relations = __get_relations_from_events(events, target_object_type)
    del events[ocel_constants.DEFAULT_OBJECT_ID]

    events = events.sort_values(
        [
            ocel_constants.DEFAULT_EVENT_TIMESTAMP,
            ocel_constants.DEFAULT_EVENT_ID,
        ]
    )
    relations = relations.sort_values(
        [
            ocel_constants.DEFAULT_EVENT_TIMESTAMP,
            ocel_constants.DEFAULT_EVENT_ID,
        ]
    )

    return OCEL(events=events, objects=objects, relations=relations)


def from_interleavings(
    df1: pd.DataFrame,
    df2: pd.DataFrame,
    interleavings: pd.DataFrame,
    parameters: Optional[Dict[Any, Any]] = None,
) -> OCEL:
    """
    Transforms a couple of dataframes, along with the interleavings between them, to an OCEL

    Parameters
    -----------------
    df1
        First of the two dataframes
    df2
        Second of the two dataframes
    interleavings
        Interleavings dataframe
    parameters
        Parameters of the algorithm, including:
        - Parameters.ACTIVITY_KEY => the attribute to use as activity
        - Parameters.TIMESTAMP_KEY => the attribute to use as timestamp
        - Parameters.CASE_ID_KEY => the attribute to use as case identifier
        - Parameters.CASE_ATTRIBUTE_PREFIX => the prefix identifying the attributes at the case level
        - Parameters.TARGET_OBJECT_TYPE => the name of the object type to which the cases of the first log should be mapped
        - Parameters.TARGET_OBJECT_TYPE_2 => the name of the object type to which the cases of the second log should be mapped
        - Parameters.LEFT_INDEX => the index column of the events of the first dataframe, in the interleavings dataframe
        - Parameters.RIGHT_INDEX => the index column of the events of the second dataframe, in the interleavings
                                    dataframe.
        - Parameters.DIRECTION => the direction of the interleavings (LR or RL)

    Returns
    -----------------
    ocel
        OCEL (equivalent to the provided event log)
    """
    if parameters is None:
        parameters = {}

    activity_key = exec_utils.get_param_value(
        Parameters.ACTIVITY_KEY, parameters, xes_constants.DEFAULT_NAME_KEY
    )
    timestamp_key = exec_utils.get_param_value(
        Parameters.TIMESTAMP_KEY,
        parameters,
        xes_constants.DEFAULT_TIMESTAMP_KEY,
    )
    case_id_key = exec_utils.get_param_value(
        Parameters.CASE_ID_KEY, parameters, constants.CASE_CONCEPT_NAME
    )
    case_attribute_prefix = exec_utils.get_param_value(
        Parameters.CASE_ATTRIBUTE_PREFIX,
        parameters,
        constants.CASE_ATTRIBUTE_PREFIX,
    )
    target_object_type = exec_utils.get_param_value(
        Parameters.TARGET_OBJECT_TYPE, parameters, "OTYPE"
    )
    target_object_type_2 = exec_utils.get_param_value(
        Parameters.TARGET_OBJECT_TYPE_2, parameters, "OTYPE2"
    )
    left_index = exec_utils.get_param_value(
        Parameters.LEFT_INDEX, parameters, "@@left_index"
    )
    right_index = exec_utils.get_param_value(
        Parameters.RIGHT_INDEX, parameters, "@@right_index"
    )
    direction = exec_utils.get_param_value(
        Parameters.DIRECTION, parameters, "@@direction"
    )

    events1 = __get_events_dataframe(
        df1,
        activity_key,
        timestamp_key,
        case_id_key,
        case_attribute_prefix,
        events_prefix="E1_",
    )
    objects1 = __get_objects_dataframe(
        df1, case_id_key, case_attribute_prefix, target_object_type
    )
    relations1 = __get_relations_from_events(events1, target_object_type)

    relations1_minimal = relations1[
        {
            ocel_constants.DEFAULT_EVENT_ID,
            ocel_constants.DEFAULT_OBJECT_ID,
            ocel_constants.DEFAULT_OBJECT_TYPE,
        }
    ]

    events2 = __get_events_dataframe(
        df2,
        activity_key,
        timestamp_key,
        case_id_key,
        case_attribute_prefix,
        events_prefix="E2_",
    )
    objects2 = __get_objects_dataframe(
        df2, case_id_key, case_attribute_prefix, target_object_type_2
    )
    relations2 = __get_relations_from_events(events2, target_object_type_2)
    relations2_minimal = relations2[
        {
            ocel_constants.DEFAULT_EVENT_ID,
            ocel_constants.DEFAULT_OBJECT_ID,
            ocel_constants.DEFAULT_OBJECT_TYPE,
        }
    ]

    interleavings[left_index] = "E1_" + interleavings[left_index].astype(
        int
    ).astype(str)
    interleavings[right_index] = "E2_" + interleavings[right_index].astype(
        int
    ).astype(str)
    interleavings_lr = interleavings[interleavings[direction] == "LR"][
        [left_index, right_index]
    ]
    interleavings_rl = interleavings[interleavings[direction] == "RL"][
        [left_index, right_index]
    ]

    relations3 = events1.merge(
        interleavings_lr,
        left_on=ocel_constants.DEFAULT_EVENT_ID,
        right_on=left_index,
    )
    relations3 = relations3.merge(
        relations2_minimal,
        left_on=right_index,
        right_on=ocel_constants.DEFAULT_EVENT_ID,
        suffixes=("", "_@#@#RIGHT"),
    )
    relations3[ocel_constants.DEFAULT_OBJECT_ID] = relations3[
        ocel_constants.DEFAULT_OBJECT_ID + "_@#@#RIGHT"
    ]
    relations3[ocel_constants.DEFAULT_OBJECT_TYPE] = target_object_type_2

    relations4 = events2.merge(
        interleavings_rl,
        left_on=ocel_constants.DEFAULT_EVENT_ID,
        right_on=right_index,
    )
    relations4 = relations4.merge(
        relations1_minimal,
        left_on=left_index,
        right_on=ocel_constants.DEFAULT_EVENT_ID,
        suffixes=("", "_@#@#LEFT"),
    )
    relations4[ocel_constants.DEFAULT_OBJECT_ID] = relations4[
        ocel_constants.DEFAULT_OBJECT_ID + "_@#@#LEFT"
    ]
    relations4[ocel_constants.DEFAULT_OBJECT_TYPE] = target_object_type

    del events1[ocel_constants.DEFAULT_OBJECT_ID]
    del events2[ocel_constants.DEFAULT_OBJECT_ID]

    events = pandas_utils.concat([events1, events2])
    objects = pandas_utils.concat([objects1, objects2])
    relations = pandas_utils.concat(
        [relations1, relations2, relations3, relations4]
    )

    events = events.sort_values(
        [
            ocel_constants.DEFAULT_EVENT_TIMESTAMP,
            ocel_constants.DEFAULT_EVENT_ID,
        ]
    )
    relations = relations.sort_values(
        [
            ocel_constants.DEFAULT_EVENT_TIMESTAMP,
            ocel_constants.DEFAULT_EVENT_ID,
        ]
    )

    return OCEL(events=events, objects=objects, relations=relations)


def log_to_ocel_multiple_obj_types(
    log_obj: Union[EventLog, EventStream, pd.DataFrame],
    activity_column: str,
    timestamp_column: str,
    obj_types: Collection[str],
    obj_separator: str = " AND ",
    additional_event_attributes: Optional[Collection[str]] = None,
    additional_object_attributes: Optional[Dict[str, Collection[str]]] = None,
) -> OCEL:
    """
    Converts an event log to an object-centric event log with one or more than one
    object types.

    Parameters
    ---------------
    log_obj
        Log object
    activity_column
        Activity column
    timestamp_column
        Timestamp column
    object_types
        List of columns to consider as object types
    obj_separator
        Separator between different objects in the same column
    additional_event_attributes
        Additional attributes to be considered as event attributes in the OCEL
    additional_object_attributes
        Additional attributes per object type to be considered as object attributes in the OCEL
        (dictionary in which object types are associated to their attributes, i.e.,
        {"order": ["quantity", "cost"], "invoice": ["date", "due date"]})

    Returns
    ----------------
    ocel
        Object-centric event log
    """
    log_obj = log_converter.apply(
        log_obj, variant=log_converter.Variants.TO_DATA_FRAME
    )

    if additional_event_attributes is None:
        additional_event_attributes = {}

    if additional_object_attributes is None:
        additional_object_attributes = {}

    events = []
    objects = []
    relations = []

    obj_ids = set()

    stream = log_obj.to_dict("records")
    stream = __postprocess_stream(stream)

    for index, eve in enumerate(stream):
        ocel_eve = {
            ocel_constants.DEFAULT_EVENT_ID: str(index),
            ocel_constants.DEFAULT_EVENT_ACTIVITY: eve[activity_column],
            ocel_constants.DEFAULT_EVENT_TIMESTAMP: eve[timestamp_column],
        }
        for attr in additional_event_attributes:
            if attr in eve:
                ocel_eve[attr] = eve[attr]
        events.append(ocel_eve)

        for ot in obj_types:
            try:
                objs = eve[ot].split(obj_separator)

                for obj in objs:
                    if len(obj.strip()) > 0:
                        if obj not in obj_ids:
                            obj_ids.add(obj)
                            obj_instance = {
                                ocel_constants.DEFAULT_OBJECT_ID: obj,
                                ocel_constants.DEFAULT_OBJECT_TYPE: ot,
                            }

                            if ot in additional_object_attributes:
                                for objattname in additional_object_attributes[
                                    ot
                                ]:
                                    if objattname in eve:
                                        objattvalue = eve[objattname]
                                        obj_instance[objattname] = objattvalue

                            objects.append(obj_instance)

                        rel = copy(ocel_eve)
                        rel[ocel_constants.DEFAULT_OBJECT_ID] = obj
                        rel[ocel_constants.DEFAULT_OBJECT_TYPE] = ot

                        relations.append(rel)
            except BaseException:
                pass

    events = pandas_utils.instantiate_dataframe(events)
    objects = pandas_utils.instantiate_dataframe(objects)
    relations = pandas_utils.instantiate_dataframe(relations)
    relations = relations.drop_duplicates(
        subset=[
            ocel_constants.DEFAULT_EVENT_ID,
            ocel_constants.DEFAULT_OBJECT_ID,
        ]
    )

    ocel = OCEL(events=events, objects=objects, relations=relations)
    ocel = ocel_consistency.apply(ocel)

    return ocel



'''
    PM4Py â€“ A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschrÃ¤nkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from pm4py.objects.bpmn.obj import BPMN
from typing import Optional, Dict, Any, Tuple, List
from copy import copy
import tempfile


def _get_node_center(layout, node) -> Tuple[float, float]:
    """Get the center point of a node."""
    node_layout = layout.get(node)
    x = node_layout.get_x() + node_layout.get_width() / 2
    y = node_layout.get_y() + node_layout.get_height() / 2
    return (x, y)


def _get_node_boundary_point(layout, node, target_point: Tuple[float, float]) -> Tuple[float, float]:
    """
    Calculate the intersection point on the node boundary
    towards a target point.
    """
    node_layout = layout.get(node)
    x = node_layout.get_x()
    y = node_layout.get_y()
    w = node_layout.get_width()
    h = node_layout.get_height()

    cx, cy = x + w / 2, y + h / 2
    tx, ty = target_point

    dx = tx - cx
    dy = ty - cy

    if dx == 0 and dy == 0:
        return (cx, y)  # Default to top edge

    # Calculate intersection with rectangle boundary
    if dx == 0:
        return (cx, y if dy < 0 else y + h)
    if dy == 0:
        return (x if dx < 0 else x + w, cy)

    # Check intersection with each edge
    scale_x = (w / 2) / abs(dx)
    scale_y = (h / 2) / abs(dy)
    scale = min(scale_x, scale_y)

    return (cx + dx * scale, cy + dy * scale)


def apply(
        bpmn_graph: BPMN, parameters: Optional[Dict[Any, Any]] = None
) -> BPMN:
    """
    Layouts the BPMN graph using node positioning and edge waypoints from Graphviz SVG.
    Ensures edges connect properly to node boundaries.
    """
    if parameters is None:
        parameters = {}

    from pm4py.visualization.bpmn.variants import classic as bpmn_visualizer
    from pm4py.visualization.common import svg_pos_parser

    layout = bpmn_graph.get_layout()

    filename_svg = tempfile.NamedTemporaryFile(suffix=".svg", delete=False)
    filename_svg.close()

    vis_parameters = copy(parameters)
    vis_parameters["format"] = "svg"
    vis_parameters["include_name_in_events"] = False
    vis_parameters["endpoints_shape"] = "box"

    gviz = bpmn_visualizer.apply(bpmn_graph, parameters=vis_parameters)
    bpmn_visualizer.save(gviz, filename_svg.name)

    nodes_p, edges_p = svg_pos_parser.apply(filename_svg.name)

    # First pass: layout all nodes
    for node in list(bpmn_graph.get_nodes()):
        node_id = str(id(node))
        if node_id in nodes_p:
            node_info = nodes_p[node_id]
            if node_info["polygon"] is not None:
                min_x = min(p[0] for p in node_info["polygon"])
                max_x = max(p[0] for p in node_info["polygon"])
                min_y = min(p[1] for p in node_info["polygon"])
                max_y = max(p[1] for p in node_info["polygon"])

                layout.get(node).set_width(max_x - min_x)
                layout.get(node).set_height(max_y - min_y)
                layout.get(node).set_x(min_x)
                layout.get(node).set_y(min_y)

    # Second pass: layout edges with proper connection points
    for flow in list(bpmn_graph.get_flows()):
        flow_id = (str(id(flow.source)), str(id(flow.target)))
        if flow_id in edges_p:
            flow_info = edges_p[flow_id]
            if flow_info["waypoints"] is not None and len(flow_info["waypoints"]) >= 2:
                flow.del_waypoints()

                waypoints = list(flow_info["waypoints"])

                # Adjust start point to source node boundary
                if len(waypoints) > 1:
                    start_boundary = _get_node_boundary_point(layout, flow.source, waypoints[1])
                    waypoints[0] = start_boundary

                # Adjust end point to target node boundary
                if len(waypoints) > 1:
                    end_boundary = _get_node_boundary_point(layout, flow.target, waypoints[-2])
                    waypoints[-1] = end_boundary

                for wayp in waypoints:
                    flow.add_waypoint(wayp)

    return bpmn_graph



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
import numpy as np
from pm4py.util.business_hours import BusinessHours
from pm4py.util import constants


def get_class_representation_by_str_ev_attr_value_presence(
    log, str_attr_name, str_attr_value
):
    """
    Get the representation for the target part of the decision tree learning
    if the focus is on the presence of a given value of a (string) event attribute

    Parameters
    -------------
    log
        Trace log
    str_attr_name
        Attribute name to consider
    str_attr_value
        Attribute value to consider

    Returns
    -------------
    target
        Target part for decision tree learning
    classes
        Name of the classes, in order
    """
    count = 0
    dictionary = {}
    target = []
    classes = []

    for trace in log:
        value = False
        for event in trace:
            if (
                str_attr_name in event
                and event[str_attr_name] == str_attr_value
            ):
                value = True
        if not str(value) in dictionary:
            dictionary[str(value)] = count
            classes.append(str(value))
            count = count + 1
        target.append(dictionary[str(value)])

    target = np.array(target)
    return target, classes


def get_class_representation_by_str_ev_attr_value_value(log, str_attr_name):
    """
    Get the representation for the target part of the decision tree learning
    if the focus is on all (string) values of an event attribute

    Parameters
    ------------
    log
        Trace log
    str_attr_name
        Attribute name to consider

    Returns
    ------------
    target
        Target part for decision tree learning
    classes
        Name of the classes, in order
    """
    count = 0
    dictionary = {}
    target = []
    classes = []

    for trace in log:
        value = "UNDEFINED"
        for event in trace:
            if str_attr_name in event and event[str_attr_name]:
                value = event[str_attr_name]
        if not str(value) in dictionary:
            dictionary[str(value)] = count
            classes.append(str(value))
            count = count + 1
        target.append(dictionary[str(value)])

    target = np.array(target)
    return target, classes


def get_class_representation_by_trace_duration(
    log, target_trace_duration, timestamp_key="time:timestamp", parameters=None
):
    """
    Get class representation by splitting traces according to trace duration

    Parameters
    ------------
    log
        Trace log
    target_trace_duration
        Target trace duration
    timestamp_key
        Timestamp key

    Returns
    ------------
    target
        Target part for decision tree learning
    classes
        Name of the classes, in order
    """
    if parameters is None:
        parameters = {}

    business_hours = (
        parameters["business_hours"]
        if "business_hours" in parameters
        else False
    )
    business_hours_slots = (
        parameters["business_hour_slots"]
        if "business_hour_slots" in parameters
        else constants.DEFAULT_BUSINESS_HOUR_SLOTS
    )

    count = 0
    dictionary = {}
    target = []
    classes = []

    for trace in log:
        value = "LESSEQUAL"
        if (
            len(trace) > 0
            and timestamp_key in trace[0]
            and timestamp_key in trace[-1]
        ):
            timestamp_st = trace[0][timestamp_key]
            timestamp_et = trace[-1][timestamp_key]
            if business_hours:
                bh = BusinessHours(
                    timestamp_st,
                    timestamp_et,
                    business_hour_slots=business_hours_slots,
                )
                diff = bh.get_seconds()
            else:
                diff = (timestamp_et - timestamp_st).total_seconds()
            if diff > target_trace_duration:
                value = "GREATER"
        if not str(value) in dictionary:
            dictionary[str(value)] = count
            classes.append(str(value))
            count = count + 1
        target.append(dictionary[str(value)])

    target = np.array(target)
    return target, classes



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''

import os
import sqlite3
from datetime import datetime
import pandas as pd
from enum import Enum
from typing import Optional, Dict, Any
from pm4py.util import exec_utils, pandas_utils


class Parameters(Enum):
    HISTORY_DB_PATH = "history_db_path"


def apply(parameters: Optional[Dict[Any, str]] = None) -> pd.DataFrame:
    """
    Extracts a dataframe containing the navigation history of Google Chrome.
    Please keep Google Chrome history closed when extracting.

    CASE ID (case:concept:name) => an identifier of the profile that has been extracted
    ACTIVITY (concept:name) => the complete path of the website, minus the GET arguments
    TIMESTAMP (time:timestamp) => the timestamp of visit

    Parameters
    --------------
    Parameters.HISTORY_DB_PATH
        Path to the history DB path of Google Chrome (default: position of the Windows folder)

    Returns
    --------------
    dataframe
        Pandas dataframe
    """
    if parameters is None:
        parameters = {}

    history_db_path = exec_utils.get_param_value(
        Parameters.HISTORY_DB_PATH,
        parameters,
        "C:\\Users\\"
        + os.getenv("USERNAME")
        + "\\AppData\\Local\\Google\\Chrome\\User Data",
    )
    print(history_db_path)

    if os.path.isdir(history_db_path):
        profiles = [
            (os.path.join(history_db_path, x, "History"), x)
            for x in os.listdir(history_db_path)
            if x.startswith("Profile ")
        ]
    else:
        profiles = [(history_db_path, "DEFAULT")]

    profiles = [x for x in profiles if os.path.exists(x[0])]

    events = []
    for prof in profiles:
        conn = sqlite3.connect(prof[0])
        curs = conn.cursor()
        curs.execute(
            "SELECT b.url, datetime(a.visit_time / 1000000 + (strftime('%s', '1601-01-01')), 'unixepoch', 'localtime') FROM (SELECT url, visit_time FROM visits) a JOIN (SELECT id, url FROM urls) b ON a.url = b.id"
        )
        res = curs.fetchall()
        for r in res:
            ev = {
                "case:concept:name": prof[1],
                "concept:name": r[0]
                .split("//")[-1]
                .split("?")[0]
                .replace(",", ""),
                "complete_url": r[0],
                "domain": r[0].split("//")[-1].split("/")[0],
                "url_wo_parameters": r[0].split("//")[-1].split("?")[0],
                "time:timestamp": datetime.strptime(r[1], "%Y-%m-%d %H:%M:%S"),
            }
            if (
                len(ev["case:concept:name"].strip()) > 0
                and len(ev["concept:name"].strip()) > 0
            ):
                events.append(ev)
        curs.close()
        conn.close()

    dataframe = pandas_utils.instantiate_dataframe(events)
    if len(dataframe) > 0:
        dataframe = pandas_utils.insert_index(
            dataframe, "@@index", copy_dataframe=False, reset_index=False
        )
        dataframe = dataframe.sort_values(["time:timestamp", "@@index"])
        dataframe["@@case_index"] = dataframe.groupby(
            "case:concept:name", sort=False
        ).ngroup()
        dataframe = dataframe.sort_values(
            ["@@case_index", "time:timestamp", "@@index"]
        )
    return dataframe



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or 
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''

import math
import tempfile
from uuid import uuid4
from enum import Enum
from typing import Optional, Dict, Any, Union

from graphviz import Digraph

from pm4py.objects.heuristics_net.obj import HeuristicsNet
from pm4py.util import constants, exec_utils
from pm4py.util.vis_utils import human_readable_stat


class Parameters(Enum):
    FORMAT = "format"
    BGCOLOR = "bgcolor"
    ENABLE_GRAPH_TITLE = "enable_graph_title"
    GRAPH_TITLE = "graph_title"


def get_corr_hex(num):
    """
    Gets correspondence between a number
    and an hexadecimal string

    Parameters
    -------------
    num
        Number

    Returns
    -------------
    hex_string
        Hexadecimal string
    """
    if num < 10:
        return str(int(num))
    elif num < 11:
        return "A"
    elif num < 12:
        return "B"
    elif num < 13:
        return "C"
    elif num < 14:
        return "D"
    elif num < 15:
        return "E"
    elif num < 16:
        return "F"


def transform_to_hex(graycolor):
    """
    Transform color to hexadecimal representation

    Parameters
    -------------
    graycolor
        Gray color (int from 0 to 255)

    Returns
    -------------
    hex_string
        Hexadecimal color
    """
    left0 = graycolor / 16
    right0 = graycolor % 16

    left00 = get_corr_hex(left0)
    right00 = get_corr_hex(right0)

    return "#" + left00 + right00 + left00 + right00 + left00 + right00


def transform_to_hex_2(color):
    """
    Transform color to hexadecimal representation

    Parameters
    -------------
    color
        Gray color (int from 0 to 255)

    Returns
    -------------
    hex_string
        Hexadecimal color
    """
    color = 255 - color
    color2 = 255 - color

    left0 = color / 16
    right0 = color % 16

    left1 = color2 / 16
    right1 = color2 % 16

    left0 = get_corr_hex(left0)
    right0 = get_corr_hex(right0)
    left1 = get_corr_hex(left1)
    right1 = get_corr_hex(right1)

    return "#" + left0 + right0 + left1 + right1 + left1 + right1


def get_graph(
    heu_net: HeuristicsNet,
    parameters: Optional[Dict[Union[str, Parameters], Any]] = None,
) -> Digraph:
    """
    Gets a representation of an Heuristics Net

    Parameters
    -------------
    heu_net
        Heuristics net
    parameters
        Possible parameters of the algorithm, including:
            - Parameters.FORMAT

    Returns
    ------------
    graph
        Graphviz graph
    """
    if parameters is None:
        parameters = {}

    enable_graph_title = exec_utils.get_param_value(
        Parameters.ENABLE_GRAPH_TITLE,
        parameters,
        constants.DEFAULT_ENABLE_GRAPH_TITLES,
    )
    graph_title = exec_utils.get_param_value(
        Parameters.GRAPH_TITLE, parameters, "Heuristics Net"
    )

    bgcolor = exec_utils.get_param_value(
        Parameters.BGCOLOR, parameters, constants.DEFAULT_BGCOLOR
    )
    image_format = exec_utils.get_param_value(
        Parameters.FORMAT, parameters, "png"
    )

    filename = tempfile.NamedTemporaryFile(suffix=".gv")
    filename.close()

    graph = Digraph(filename=filename.name, strict=True, engine="dot")
    graph.attr(bgcolor=bgcolor)

    if enable_graph_title:
        graph.attr(
            label=graph_title,
            labelloc="top",
            labeljust="center",
            fontsize="20",
        )

    corr_nodes = {}
    corr_nodes_names = {}
    is_frequency = False

    start_end_nodes_set = set()

    for index, sa_list in enumerate(heu_net.start_activities):
        start_end_nodes_set = start_end_nodes_set.union(
            {n for n in sa_list if n in corr_nodes_names}
        )

    for index, ea_list in enumerate(heu_net.end_activities):
        start_end_nodes_set = start_end_nodes_set.union(
            {n for n in ea_list if n in corr_nodes_names}
        )

    for node_name in heu_net.nodes:
        node = heu_net.nodes[node_name]
        if (
            node_name in start_end_nodes_set
            or node.input_connections
            or node.output_connections
        ):
            node_occ = node.node_occ
            graycolor = transform_to_hex_2(
                max(255 - math.log(node_occ) * 9, 0)
            )
            node_id = str(uuid4())
            if node.node_type == "frequency":
                is_frequency = True
                node_label = f"{node_name} ({node_occ})"
            else:
                sojourn = (
                    human_readable_stat(heu_net.sojourn_times[node_name])
                    if node_name in heu_net.sojourn_times
                    else "0s"
                )
                node_label = f"{node_name} ({sojourn})"

            graph.node(
                node_id,
                label=node_label,
                shape="box",
                style="filled",
                fillcolor=node.get_fill_color(graycolor),
                fontcolor=node.get_font_color(),
            )
            corr_nodes[node] = node_id
            corr_nodes_names[node_name] = node_id

    # gets max arc value
    max_arc_value = -1
    for node_name in heu_net.nodes:
        node = heu_net.nodes[node_name]
        for other_node in node.output_connections:
            if other_node in corr_nodes:
                for edge in node.output_connections[other_node]:
                    max_arc_value = max(max_arc_value, edge.repr_value)

    for node_name in heu_net.nodes:
        node = heu_net.nodes[node_name]
        for other_node in node.output_connections:
            if other_node in corr_nodes:
                for edge in node.output_connections[other_node]:
                    this_pen_width = 1.0 + math.log(1 + edge.repr_value) / 11.0
                    repr_value = str(edge.repr_value)
                    edge_kwargs = {
                        "color": edge.get_color(),
                        "fontcolor": edge.get_font_color(),
                        "penwidth": str(edge.get_penwidth(this_pen_width)),
                    }
                    if edge.net_name:
                        if node.node_type == "frequency":
                            label = f"{edge.net_name} ({repr_value})"
                        else:
                            label = (
                                f"{edge.net_name} "
                                f"({human_readable_stat(repr_value)})"
                            )
                    else:
                        if node.node_type == "frequency":
                            label = repr_value
                        else:
                            label = human_readable_stat(repr_value)
                    graph.edge(
                        corr_nodes[node], corr_nodes[other_node], label=label, **edge_kwargs
                    )

    for index, sa_list in enumerate(heu_net.start_activities):
        effective_sa_list = [n for n in sa_list if n in corr_nodes_names]
        if effective_sa_list:
            start_id = f"start_{index}"
            graph.node(
                start_id,
                label="@@S",
                color=heu_net.default_edges_color[index],
                fontsize="8",
                fontcolor="#32CD32",
                fillcolor="#32CD32",
                style="filled",
            )
            for node_name in effective_sa_list:
                sa = corr_nodes_names[node_name]
                if type(heu_net.start_activities[index]) is dict:
                    occ = heu_net.start_activities[index][node_name]
                    if occ >= heu_net.min_dfg_occurrences:
                        if is_frequency:
                            this_pen_width = 1.0 + math.log(1 + occ) / 11.0
                            penwidth = str(this_pen_width)
                            if heu_net.net_name[index]:
                                label = f"{heu_net.net_name[index]} ({occ})"
                            else:
                                label = str(occ)
                            graph.edge(
                                start_id,
                                sa,
                                label=label,
                                color=heu_net.default_edges_color[index],
                                fontcolor=heu_net.default_edges_color[index],
                                penwidth=penwidth,
                            )
                        else:
                            graph.edge(
                                start_id,
                                sa,
                                label=heu_net.net_name[index],
                                color=heu_net.default_edges_color[index],
                                fontcolor=heu_net.default_edges_color[index],
                            )
                else:
                    graph.edge(
                        start_id,
                        sa,
                        label=heu_net.net_name[index],
                        color=heu_net.default_edges_color[index],
                        fontcolor=heu_net.default_edges_color[index],
                    )

    for index, ea_list in enumerate(heu_net.end_activities):
        effective_ea_list = [n for n in ea_list if n in corr_nodes_names]
        if effective_ea_list:
            end_id = f"end_{index}"
            graph.node(
                end_id,
                label="@@E",
                color="#FFA500",
                fillcolor="#FFA500",
                fontcolor="#FFA500",
                fontsize="8",
                style="filled",
            )
            for node_name in effective_ea_list:
                ea = corr_nodes_names[node_name]
                if type(heu_net.end_activities[index]) is dict:
                    occ = heu_net.end_activities[index][node_name]
                    if occ >= heu_net.min_dfg_occurrences:
                        if is_frequency:
                            this_pen_width = 1.0 + math.log(1 + occ) / 11.0
                            penwidth = str(this_pen_width)
                            if heu_net.net_name[index]:
                                label = f"{heu_net.net_name[index]} ({occ})"
                            else:
                                label = str(occ)
                            graph.edge(
                                ea,
                                end_id,
                                label=label,
                                color=heu_net.default_edges_color[index],
                                fontcolor=heu_net.default_edges_color[index],
                                penwidth=penwidth,
                            )
                        else:
                            graph.edge(
                                ea,
                                end_id,
                                label=heu_net.net_name[index],
                                color=heu_net.default_edges_color[index],
                                fontcolor=heu_net.default_edges_color[index],
                            )
                else:
                    graph.edge(
                        ea,
                        end_id,
                        label=heu_net.net_name[index],
                        color=heu_net.default_edges_color[index],
                        fontcolor=heu_net.default_edges_color[index],
                    )

    graph.format = image_format.replace("html", "plain-ext")

    return graph


def apply(
    heu_net: HeuristicsNet,
    parameters: Optional[Dict[Union[str, Parameters], Any]] = None,
) -> str:
    """
    Gets a representation of an Heuristics Net

    Parameters
    -------------
    heu_net
        Heuristics net
    parameters
        Possible parameters of the algorithm, including:
            - Parameters.FORMAT

    Returns
    ------------
    gviz
        Representation of the Heuristics Net
    """
    if parameters is None:
        parameters = {}


    graph = get_graph(heu_net, parameters=parameters)

    return graph



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''

import uuid
from collections import Counter
from copy import deepcopy
from enum import Enum

from pm4py.objects.bpmn.obj import BPMN
from pm4py.objects.bpmn.util.sorting import get_sorted_nodes_edges
from pm4py.util import exec_utils
import tempfile


class EndpointDirection(Enum):
    RIGHT = "right"
    LEFT = "left"
    TOP = "top"
    BOTTOM = "bottom"


class Parameters(Enum):
    TASK_WH = "task_wh"
    SCREEN_SIZE_X = "screen_size_x"
    SCREEN_SIZE_Y = "screen_size_y"
    SCALING_FACTOR = "scaling_factor"


def get_right_edge_coord(layout, node, p, partial_counter, total_counter):
    y_factor = partial_counter[EndpointDirection.RIGHT] / (
        total_counter[EndpointDirection.RIGHT] + 1.0
    )
    new_x = p[0] + layout.get(node).get_width()
    new_y = p[1] + round(layout.get(node).get_height() * y_factor)
    return (new_x, new_y)


def get_left_edge_coord(layout, node, p, partial_counter, total_counter):
    y_factor = partial_counter[EndpointDirection.LEFT] / (
        total_counter[EndpointDirection.LEFT] + 1.0
    )
    new_x = p[0]
    new_y = p[1] + round(layout.get(node).get_height() * y_factor)
    return (new_x, new_y)


def get_top_edge_coord(layout, node, p, partial_counter, total_counter):
    x_factor = partial_counter[EndpointDirection.TOP] / (
        total_counter[EndpointDirection.TOP] + 1.0
    )
    new_x = p[0] + round(layout.get(node).get_width() * x_factor)
    new_y = p[1]
    return (new_x, new_y)


def get_bottom_edge_coord(layout, node, p, partial_counter, total_counter):
    x_factor = partial_counter[EndpointDirection.BOTTOM] / (
        total_counter[EndpointDirection.BOTTOM] + 1.0
    )
    new_x = p[0] + round(layout.get(node).get_width() * x_factor)
    new_y = p[1] + layout.get(node).get_height()
    return (new_x, new_y)


def apply(bpmn_graph, parameters=None):
    """
    Layouts a BPMN graph (inserting the positions of the nodes and the layouting of the edges)

    Parameters
    -------------
    bpmn_graph
        BPMN graph
    parameters
        Parameters of the algorithm:
        - Parameters.SCREEN_SIZE_X => target size of the screen in pixels (default 1920.0)
        - Parameters.SCREEN_SIZE_Y => target size of the screen in pixels (default 1080.0)
        - Parameters.SCALING_FACTOR => scaling factor of the layouting (defualt 2.5)

    Returns
    -------------
    bpmn_graph
        BPMN graph with layout information
    """
    from graphviz import Digraph
    from pm4py.visualization.common import save as gsave
    from pm4py.visualization.common import svg_pos_parser

    if parameters is None:
        parameters = {}

    screen_size_x = exec_utils.get_param_value(
        Parameters.SCREEN_SIZE_X, parameters, 1920.0
    )
    screen_size_y = exec_utils.get_param_value(
        Parameters.SCREEN_SIZE_Y, parameters, 1080.0
    )
    scaling_factor = exec_utils.get_param_value(
        Parameters.SCALING_FACTOR, parameters, 2.5
    )

    nodes = bpmn_graph.get_nodes()
    flows = bpmn_graph.get_flows()
    layout = bpmn_graph.get_layout()

    filename_gv = tempfile.NamedTemporaryFile(suffix=".gv")
    filename_gv.close()
    filename_svg = tempfile.NamedTemporaryFile(suffix=".svg")
    filename_svg.close()
    viz = Digraph(
        bpmn_graph.get_name(), filename=filename_gv.name, engine="dot"
    )
    viz.format = "svg"
    viz.graph_attr["rankdir"] = "LR"

    graph_nodes, graph_edges = get_sorted_nodes_edges(bpmn_graph)

    nodes_dict = {}
    inv_nodes_dict = {}
    for n in graph_nodes:
        node_uuid = str(uuid.uuid4()).replace("-", "")
        nodes_dict[n] = node_uuid
        inv_nodes_dict[node_uuid] = n
        viz.node(node_uuid, label=" ", shape="box")

    for tup in graph_edges:
        viz.edge(nodes_dict[tup[0]], nodes_dict[tup[1]])

    gsave.save(viz, filename_svg.name)

    nodes_p, edges_p = svg_pos_parser.apply(filename_svg.name)

    nodes_pos = {}
    for node in nodes_p:
        nodes_pos[inv_nodes_dict[node]] = nodes_p[node]["polygon"]

    endpoints_wh = exec_utils.get_param_value(
        Parameters.TASK_WH, parameters, 30
    )
    task_wh = exec_utils.get_param_value(Parameters.TASK_WH, parameters, 60)

    # add node positions to BPMN nodes
    for n in graph_nodes:
        node_pos = nodes_pos[n][0]

        pos_x = float(node_pos[0])
        pos_y = float(node_pos[1])
        layout.get(n).set_x(pos_x)
        layout.get(n).set_y(pos_y)
        if isinstance(n, BPMN.Task):
            this_width = min(
                round(2 * task_wh),
                round(2 * (len(n.get_name()) + 7) * task_wh / 22.0),
            )
            layout.get(n).set_width(this_width)
            layout.get(n).set_height(task_wh)
        elif isinstance(n, BPMN.StartEvent) or isinstance(n, BPMN.EndEvent):
            layout.get(n).set_width(endpoints_wh)
            layout.get(n).set_height(endpoints_wh)
        else:
            layout.get(n).set_width(task_wh)
            layout.get(n).set_height(task_wh)

    max_x = max(1, max(abs(layout.get(node).get_x()) for node in nodes))
    max_y = max(1, max(abs(layout.get(node).get_y()) for node in nodes))
    different_x = len(set(layout.get(node).get_x() for node in nodes))
    different_y = len(set(layout.get(node).get_y() for node in nodes))

    stretch_fact_x = scaling_factor * 1.25 * screen_size_x / max_x
    stretch_fact_y = scaling_factor * screen_size_y / max_y

    for node in nodes:
        layout.get(node).set_x(
            round(layout.get(node).get_x() * stretch_fact_x)
        )
        layout.get(node).set_y(
            round(layout.get(node).get_y() * stretch_fact_y)
        )

    min_x = min(layout.get(node).get_x() for node in nodes)
    min_y = min(layout.get(node).get_y() for node in nodes)

    for node in nodes:
        x_coord = layout.get(node).get_x() - min_x
        y_coord = layout.get(node).get_y() - min_y
        layout.get(node).set_x(x_coord)
        layout.get(node).set_y(y_coord)

    outgoing_edges = dict()
    ingoing_edges = dict()
    sources_dict = dict()
    targets_dict = dict()

    for flow in flows:
        source = flow.get_source()
        target = flow.get_target()

        x_src = layout.get(source).get_x()
        x_trg = layout.get(target).get_x()
        y_src = layout.get(source).get_y()
        y_trg = layout.get(target).get_y()

        sources_dict[(x_src, y_src)] = source
        targets_dict[(x_trg, y_trg)] = target

        diff_x = abs(x_trg - x_src)
        diff_y = abs(y_src - y_trg)

        if not (x_src, y_src) in outgoing_edges:
            outgoing_edges[(x_src, y_src)] = {}
        outgoing_edges[(x_src, y_src)][(x_trg, y_trg)] = {
            EndpointDirection.RIGHT: 0.0,
            EndpointDirection.LEFT: 0.0,
            EndpointDirection.TOP: 0.0,
            EndpointDirection.BOTTOM: 0.0,
        }
        if not (x_trg, y_trg) in ingoing_edges:
            ingoing_edges[(x_trg, y_trg)] = {}
        ingoing_edges[(x_trg, y_trg)][(x_src, y_src)] = {
            EndpointDirection.RIGHT: 0.0,
            EndpointDirection.LEFT: 0.0,
            EndpointDirection.TOP: 0.0,
            EndpointDirection.BOTTOM: 0.0,
        }

        if x_trg > x_src:
            outgoing_edges[(x_src, y_src)][(x_trg, y_trg)][
                EndpointDirection.RIGHT
            ] = diff_x / (diff_x + diff_y)
            ingoing_edges[(x_trg, y_trg)][(x_src, y_src)][
                EndpointDirection.LEFT
            ] = diff_x / (diff_x + diff_y)
        else:
            outgoing_edges[(x_src, y_src)][(x_trg, y_trg)][
                EndpointDirection.LEFT
            ] = diff_x / (diff_x + diff_y)
            ingoing_edges[(x_trg, y_trg)][(x_src, y_src)][
                EndpointDirection.RIGHT
            ] = diff_x / (diff_x + diff_y)

        if y_src > y_trg:
            outgoing_edges[(x_src, y_src)][(x_trg, y_trg)][
                EndpointDirection.TOP
            ] = diff_y / (diff_x + diff_y)
            ingoing_edges[(x_trg, y_trg)][(x_src, y_src)][
                EndpointDirection.BOTTOM
            ] = diff_y / (diff_x + diff_y)
        else:
            outgoing_edges[(x_src, y_src)][(x_trg, y_trg)][
                EndpointDirection.BOTTOM
            ] = diff_y / (diff_x + diff_y)
            ingoing_edges[(x_trg, y_trg)][(x_src, y_src)][
                EndpointDirection.TOP
            ] = diff_y / (diff_x + diff_y)

    # normalization
    outgoing_edges0 = deepcopy(outgoing_edges)
    ingoing_edges0 = deepcopy(ingoing_edges)

    for p1 in outgoing_edges:
        sum_right = 0.0
        sum_left = 0.0
        sum_top = 0.0
        sum_bottom = 0.0
        for p2 in outgoing_edges[p1]:
            sum_right += outgoing_edges0[p1][p2][EndpointDirection.RIGHT]
            sum_left += outgoing_edges0[p1][p2][EndpointDirection.LEFT]
            sum_top += outgoing_edges0[p1][p2][EndpointDirection.TOP]
            sum_bottom += outgoing_edges0[p1][p2][EndpointDirection.BOTTOM]
        if p1 in ingoing_edges:
            for p2 in ingoing_edges[p1]:
                sum_right += ingoing_edges0[p1][p2][EndpointDirection.RIGHT]
                sum_left += ingoing_edges0[p1][p2][EndpointDirection.LEFT]
                sum_top += ingoing_edges0[p1][p2][EndpointDirection.TOP]
                sum_bottom += ingoing_edges0[p1][p2][EndpointDirection.BOTTOM]
        for p2 in outgoing_edges[p1]:
            if sum_right > 0:
                outgoing_edges[p1][p2][EndpointDirection.RIGHT] = (
                    outgoing_edges[p1][p2][EndpointDirection.RIGHT] ** 2
                    / sum_right
                )
            if sum_left > 0:
                outgoing_edges[p1][p2][EndpointDirection.LEFT] = (
                    outgoing_edges[p1][p2][EndpointDirection.LEFT] ** 2
                    / sum_left
                )
            if sum_top > 0:
                outgoing_edges[p1][p2][EndpointDirection.TOP] = (
                    outgoing_edges[p1][p2][EndpointDirection.TOP] ** 2
                    / sum_top
                )
            if sum_bottom > 0:
                outgoing_edges[p1][p2][EndpointDirection.BOTTOM] = (
                    outgoing_edges[p1][p2][EndpointDirection.BOTTOM] ** 2
                    / sum_bottom
                )
    for p1 in ingoing_edges:
        sum_right = 0.0
        sum_left = 0.0
        sum_top = 0.0
        sum_bottom = 0.0
        for p2 in ingoing_edges[p1]:
            sum_right += ingoing_edges0[p1][p2][EndpointDirection.RIGHT]
            sum_left += ingoing_edges0[p1][p2][EndpointDirection.LEFT]
            sum_top += ingoing_edges0[p1][p2][EndpointDirection.TOP]
            sum_bottom += ingoing_edges0[p1][p2][EndpointDirection.BOTTOM]
        if p1 in outgoing_edges:
            for p2 in outgoing_edges[p1]:
                sum_right += outgoing_edges0[p1][p2][EndpointDirection.RIGHT]
                sum_left += outgoing_edges0[p1][p2][EndpointDirection.LEFT]
                sum_top += outgoing_edges0[p1][p2][EndpointDirection.TOP]
                sum_bottom += outgoing_edges0[p1][p2][EndpointDirection.BOTTOM]
        for p2 in ingoing_edges[p1]:
            if sum_right > 0:
                ingoing_edges[p1][p2][EndpointDirection.RIGHT] = (
                    ingoing_edges[p1][p2][EndpointDirection.RIGHT] ** 2
                    / sum_right
                )
            if sum_left > 0:
                ingoing_edges[p1][p2][EndpointDirection.LEFT] = (
                    ingoing_edges[p1][p2][EndpointDirection.LEFT] ** 2
                    / sum_left
                )
            if sum_top > 0:
                ingoing_edges[p1][p2][EndpointDirection.TOP] = (
                    ingoing_edges[p1][p2][EndpointDirection.TOP] ** 2 / sum_top
                )
            if sum_bottom > 0:
                ingoing_edges[p1][p2][EndpointDirection.BOTTOM] = (
                    ingoing_edges[p1][p2][EndpointDirection.BOTTOM] ** 2
                    / sum_bottom
                )

    # keep best direction
    for p1 in outgoing_edges:
        for p2 in outgoing_edges[p1]:
            vals = sorted(
                [(x, y) for x, y in outgoing_edges[p1][p2].items()],
                key=lambda x: x[1],
                reverse=True,
            )
            outgoing_edges[p1][p2] = vals[0][0]
    for p1 in ingoing_edges:
        for p2 in ingoing_edges[p1]:
            vals = sorted(
                [(x, y) for x, y in ingoing_edges[p1][p2].items()],
                key=lambda x: x[1],
                reverse=True,
            )
            ingoing_edges[p1][p2] = vals[0][0]

    total_counter = dict()
    partial_counter = dict()
    for p1 in outgoing_edges:
        if p1 not in total_counter:
            total_counter[p1] = Counter()
        for p2 in outgoing_edges[p1]:
            dir = outgoing_edges[p1][p2]
            total_counter[p1][dir] += 1
    for p1 in ingoing_edges:
        if p1 not in total_counter:
            total_counter[p1] = Counter()
        for p2 in ingoing_edges[p1]:
            dir = ingoing_edges[p1][p2]
            total_counter[p1][dir] += 1

    outgoing_edges_dirs = deepcopy(outgoing_edges)
    ingoing_edges_dirs = deepcopy(ingoing_edges)

    # decide exiting/entering point for edges
    for p1 in outgoing_edges:
        node = sources_dict[p1]
        if p1 not in partial_counter:
            partial_counter[p1] = Counter()
        sorted_outgoing_edges = sorted(
            outgoing_edges[p1], key=lambda x: x, reverse=False
        )
        for p2 in sorted_outgoing_edges:
            dir = outgoing_edges[p1][p2]
            partial_counter[p1][dir] += 1
            if dir == EndpointDirection.RIGHT:
                outgoing_edges[p1][p2] = get_right_edge_coord(
                    layout, node, p1, partial_counter[p1], total_counter[p1]
                )
            elif dir == EndpointDirection.LEFT:
                outgoing_edges[p1][p2] = get_left_edge_coord(
                    layout, node, p1, partial_counter[p1], total_counter[p1]
                )
            elif dir == EndpointDirection.TOP:
                outgoing_edges[p1][p2] = get_top_edge_coord(
                    layout, node, p1, partial_counter[p1], total_counter[p1]
                )
            elif dir == EndpointDirection.BOTTOM:
                outgoing_edges[p1][p2] = get_bottom_edge_coord(
                    layout, node, p1, partial_counter[p1], total_counter[p1]
                )
    for p1 in ingoing_edges:
        node = targets_dict[p1]
        if p1 not in partial_counter:
            partial_counter[p1] = Counter()
        sorted_ingoing_edges = sorted(
            ingoing_edges[p1], key=lambda x: x, reverse=False
        )
        for p2 in sorted_ingoing_edges:
            dir = ingoing_edges[p1][p2]
            partial_counter[p1][dir] += 1
            if dir == EndpointDirection.RIGHT:
                ingoing_edges[p1][p2] = get_right_edge_coord(
                    layout, node, p1, partial_counter[p1], total_counter[p1]
                )
            elif dir == EndpointDirection.LEFT:
                ingoing_edges[p1][p2] = get_left_edge_coord(
                    layout, node, p1, partial_counter[p1], total_counter[p1]
                )
            elif dir == EndpointDirection.TOP:
                ingoing_edges[p1][p2] = get_top_edge_coord(
                    layout, node, p1, partial_counter[p1], total_counter[p1]
                )
            elif dir == EndpointDirection.BOTTOM:
                ingoing_edges[p1][p2] = get_bottom_edge_coord(
                    layout, node, p1, partial_counter[p1], total_counter[p1]
                )

    # order the left-entering ingoing edges better
    for p1 in ingoing_edges:
        vals = [(x, y) for x, y in ingoing_edges[p1].items() if y[0] == p1[0]]
        if len(vals) > 1:
            vals_x = [x[0] for x in vals]
            vals_y = [x[1] for x in vals]
            vals_x = sorted(vals_x)
            vals_y = sorted(vals_y)
            for i in range(len(vals_x)):
                ingoing_edges[p1][vals_x[i]] = vals_y[i]

    # set waypoints for edges
    for flow in flows:
        source = flow.get_source()
        target = flow.get_target()

        flow.del_waypoints()

        x_src = layout.get(source).get_x()
        x_trg = layout.get(target).get_x()
        y_src = layout.get(source).get_y()
        y_trg = layout.get(target).get_y()
        p1 = (x_src, y_src)
        p2 = (x_trg, y_trg)

        source_x = outgoing_edges[p1][p2][0]
        source_y = outgoing_edges[p1][p2][1]
        target_x = ingoing_edges[p2][p1][0]
        target_y = ingoing_edges[p2][p1][1]
        dir_source = outgoing_edges_dirs[p1][p2]
        dir_target = ingoing_edges_dirs[p2][p1]

        middle_x = (source_x + target_x) / 2.0
        middle_y = (source_y + target_y) / 2.0

        flow.add_waypoint((source_x, source_y))
        if dir_source in [EndpointDirection.LEFT, EndpointDirection.RIGHT]:
            if dir_target in [EndpointDirection.LEFT, EndpointDirection.RIGHT]:
                flow.add_waypoint((middle_x, source_y))
                flow.add_waypoint((middle_x, target_y))
            elif dir_target in [
                EndpointDirection.TOP,
                EndpointDirection.BOTTOM,
            ]:
                flow.add_waypoint((target_x, source_y))
        elif dir_source in [EndpointDirection.TOP, EndpointDirection.BOTTOM]:
            if dir_target in [EndpointDirection.TOP, EndpointDirection.BOTTOM]:
                flow.add_waypoint((source_x, middle_y))
                flow.add_waypoint((target_x, middle_y))
            elif dir_target in [
                EndpointDirection.LEFT,
                EndpointDirection.RIGHT,
            ]:
                flow.add_waypoint((source_x, target_y))

        flow.add_waypoint((target_x, target_y))

    return bpmn_graph



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''

from pm4py.objects.petri_net.obj import PetriNet
from typing import Optional, Dict, Any


def apply(
    petri_net: PetriNet, parameters: Optional[Dict[Any, Any]] = None
) -> float:
    """
    Computes the extended Cardoso metric as described in the paper:

    "Complexity Metrics for Workflow Nets"
    Lassen, Kristian Bisgaard, and Wil MP van der Aalst

    Parameters
    -------------
    petri_net
        Petri net

    Returns
    -------------
    ext_cardoso_metric
        Extended Cardoso metric
    """
    if parameters is None:
        parameters = {}

    ext_card = 0

    for place in petri_net.places:
        targets = set()
        for out_arc in place.out_arcs:
            targets1 = set()
            for out_arc2 in out_arc.target.out_arcs:
                targets1.add(out_arc2.target.name)
            targets.add(tuple(sorted(list(targets1))))
        ext_card += len(targets)

    return ext_card



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''

from pm4py.statistics.attributes.log import get as attr_get
from pm4py.objects.dfg.utils import dfg_utils
from pm4py.util import xes_constants as xes
from pm4py.util import exec_utils
from pm4py.statistics.service_time.log import get as serv_time_get
from enum import Enum
from pm4py.util import constants
from typing import Optional, Dict, Any, Tuple
import graphviz
from pm4py.objects.log.obj import EventLog
from collections import Counter
from pm4py.visualization.dfg.util import dfg_gviz


class Parameters(Enum):
    ACTIVITY_KEY = constants.PARAMETER_CONSTANT_ACTIVITY_KEY
    FORMAT = "format"
    MAX_NO_EDGES_IN_DIAGRAM = "maxNoOfEdgesInDiagram"
    START_ACTIVITIES = "start_activities"
    END_ACTIVITIES = "end_activities"
    TIMESTAMP_KEY = constants.PARAMETER_CONSTANT_TIMESTAMP_KEY
    START_TIMESTAMP_KEY = constants.PARAMETER_CONSTANT_START_TIMESTAMP_KEY
    FONT_SIZE = "font_size"
    RANKDIR = "rankdir"
    BGCOLOR = "bgcolor"
    ENABLE_GRAPH_TITLE = "enable_graph_title"
    GRAPH_TITLE = "graph_title"


def apply(
    dfg: Dict[Tuple[str, str], int],
    log: EventLog = None,
    parameters: Optional[Dict[Any, Any]] = None,
    activities_count: Dict[str, int] = None,
    serv_time: Dict[str, float] = None,
) -> graphviz.Digraph:
    """
    Visualize a frequency directly-follows graph

    Parameters
    -----------------
    dfg
        Frequency Directly-follows graph
    log
        (if provided) Event log for the calculation of statistics
    activities_count
        (if provided) Dictionary associating to each activity the number of occurrences in the log.
    serv_time
        (if provided) Dictionary associating to each activity the average service time
    parameters
        Variant-specific parameters

    Returns
    -----------------
    gviz
        Graphviz digraph
    """
    if parameters is None:
        parameters = {}

    activity_key = exec_utils.get_param_value(
        Parameters.ACTIVITY_KEY, parameters, xes.DEFAULT_NAME_KEY
    )
    image_format = exec_utils.get_param_value(
        Parameters.FORMAT, parameters, "png"
    )
    max_no_of_edges_in_diagram = exec_utils.get_param_value(
        Parameters.MAX_NO_EDGES_IN_DIAGRAM, parameters, 100000
    )
    start_activities = exec_utils.get_param_value(
        Parameters.START_ACTIVITIES, parameters, {}
    )
    end_activities = exec_utils.get_param_value(
        Parameters.END_ACTIVITIES, parameters, {}
    )
    font_size = exec_utils.get_param_value(
        Parameters.FONT_SIZE, parameters, 12
    )
    font_size = str(font_size)

    if start_activities is None:
        start_activities = dict()
    if end_activities is None:
        end_activities = dict()
    activities = sorted(
        list(
            set(dfg_utils.get_activities_from_dfg(dfg))
            .union(set(start_activities))
            .union(set(end_activities))
        )
    )

    rankdir = exec_utils.get_param_value(
        Parameters.RANKDIR, parameters, constants.DEFAULT_RANKDIR_GVIZ
    )
    bgcolor = exec_utils.get_param_value(
        Parameters.BGCOLOR, parameters, constants.DEFAULT_BGCOLOR
    )
    enable_graph_title = exec_utils.get_param_value(
        Parameters.ENABLE_GRAPH_TITLE,
        parameters,
        constants.DEFAULT_ENABLE_GRAPH_TITLES,
    )
    graph_title = exec_utils.get_param_value(
        Parameters.GRAPH_TITLE, parameters, "Frequency Directly-Follows Graph"
    )

    if activities_count is None:
        if log is not None:
            activities_count = attr_get.get_attribute_values(
                log, activity_key, parameters=parameters
            )
        else:
            # the frequency of an activity in the log is at least the number of occurrences of
            # incoming arcs in the DFG.
            # if the frequency of the start activities nodes is also provided,
            # use also that.
            activities_count = Counter({key: 0 for key in activities})
            for el in dfg:
                activities_count[el[1]] += dfg[el]
            if isinstance(start_activities, dict):
                for act in start_activities:
                    activities_count[act] += start_activities[act]

    if serv_time is None:
        if log is not None:
            serv_time = serv_time_get.apply(log, parameters=parameters)
        else:
            serv_time = {key: 0 for key in activities}

    return dfg_gviz.graphviz_visualization(
        activities_count,
        dfg,
        image_format=image_format,
        measure="frequency",
        max_no_of_edges_in_diagram=max_no_of_edges_in_diagram,
        start_activities=start_activities,
        end_activities=end_activities,
        serv_time=serv_time,
        font_size=font_size,
        bgcolor=bgcolor,
        rankdir=rankdir,
        enable_graph_title=enable_graph_title,
        graph_title=graph_title,
    )



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
import os
from abc import abstractmethod, ABC
from typing import Optional, Tuple, List, TypeVar, Generic, Dict, Any

from pm4py.algo.discovery.inductive.base_case.factory import BaseCaseFactory
from pm4py.algo.discovery.inductive.cuts.factory import CutFactory
from pm4py.algo.discovery.inductive.dtypes.im_ds import IMDataStructure
from pm4py.algo.discovery.inductive.fall_through.factory import (
    FallThroughFactory,
)
from pm4py.algo.discovery.inductive.variants.instances import IMInstance
from pm4py.objects.process_tree.obj import ProcessTree
from enum import Enum
from pm4py.util import exec_utils, constants


T = TypeVar("T", bound=IMDataStructure)


class Parameters(Enum):
    MULTIPROCESSING = "multiprocessing"


class InductiveMinerFramework(ABC, Generic[T]):
    """
    Base Class Implementing the Inductive Miner Framework.
    How to Extend:
    1. Create a dedicated IMDataStructure class (see pm4py.algo.discovery.inductive.dtypes.im_ds.py)
    2. Create dedicated Base Cases, Cuts and Fall Throughs for the newly constructed IMDataStructure
    3. Extend the BaseCaseFactory, CutFactory and FallThroughFactory with the newly created functions
    4. Create a subclass of this class indicating the type on which it is defined and the corresponding IMInstance.
    """

    def __init__(self, parameters: Optional[Dict[str, Any]] = None):
        if parameters is None:
            parameters = {}

        enable_multiprocessing = exec_utils.get_param_value(
            Parameters.MULTIPROCESSING,
            parameters,
            constants.ENABLE_MULTIPROCESSING_DEFAULT,
        )

        if enable_multiprocessing:
            from multiprocessing import Pool, Manager

            self._pool = Pool(os.cpu_count() - 1)
            self._manager = Manager()
            self._manager.support_list = []
        else:
            self._pool = None
            self._manager = None

    def apply_base_cases(
        self, obj: T, parameters: Optional[Dict[str, Any]] = None
    ) -> Optional[ProcessTree]:
        return BaseCaseFactory.apply_base_cases(
            obj, self.instance(), parameters=parameters
        )

    def find_cut(
        self, obj: T, parameters: Optional[Dict[str, Any]] = None
    ) -> Optional[Tuple[ProcessTree, List[T]]]:
        return CutFactory.find_cut(obj, self.instance(), parameters=parameters)

    def fall_through(
        self, obj: T, parameters: Optional[Dict[str, Any]] = None
    ) -> Tuple[ProcessTree, List[T]]:
        return FallThroughFactory.fall_through(
            obj,
            self.instance(),
            self._pool,
            self._manager,
            parameters=parameters,
        )

    def apply(
        self, obj: T, parameters: Optional[Dict[str, Any]] = None
    ) -> ProcessTree:
        tree = self.apply_base_cases(obj, parameters)
        if tree is None:
            cut = self.find_cut(obj, parameters)
            if cut is not None:
                tree = self._recurse(cut[0], cut[1], parameters=parameters)
        if tree is None:
            ft = self.fall_through(obj, parameters)
            tree = self._recurse(ft[0], ft[1], parameters=parameters)
        return tree

    def _recurse(
        self,
        tree: ProcessTree,
        objs: List[T],
        parameters: Optional[Dict[str, Any]] = None,
    ):
        children = [self.apply(obj, parameters=parameters) for obj in objs]
        for c in children:
            c.parent = tree
        tree.children.extend(children)
        return tree

    @abstractmethod
    def instance(self) -> IMInstance:
        pass



'''
    PM4Py â€“ A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschrÃ¤nkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from pm4py.objects.petri_net.utils.petri_utils import (
    remove_arc,
    remove_transition,
    remove_place,
    add_arc_from_to,
    pre_set,
    post_set,
    get_arc_type,
)
from pm4py.objects.petri_net.obj import PetriNet
from pm4py.objects.petri_net import properties
import itertools
from itertools import combinations, chain


######################################################################
# NEW HELPER FUNCTIONS
######################################################################

def _has_only_this_transition_as_output(place, transition):
    """
    Checks if all arcs exiting the given place go to exactly this 'transition'
    (i.e., place -> transition, no other transitions).
    """
    outgoing_transitions = {
        arc.target for arc in place.out_arcs if isinstance(arc.target, PetriNet.Transition)
    }
    return outgoing_transitions == {transition}


def _has_only_this_transition_as_input(place, transition):
    """
    Checks if all arcs entering the given place come from exactly this 'transition'
    (i.e., transition -> place, no other transitions).
    """
    incoming_transitions = {
        arc.source for arc in place.in_arcs if isinstance(arc.source, PetriNet.Transition)
    }
    return incoming_transitions == {transition}


######################################################################
# REFACTOR: SINGLE-ENTRY / SINGLE-EXIT REDUCTIONS
######################################################################

def reduce_single_entry_transitions(net):
    """
    Reduces the number of single-entry silent transitions in the Petri net
    by merging them upward when the place feeding them does not feed any
    other transition.

    Parameters
    ----------------
    net
        Petri net
    """
    cont = True
    while cont:
        cont = False
        # Identify silent transitions with exactly 1 input place.
        single_entry_transitions = [
            t for t in net.transitions
            if t.label is None and len(t.in_arcs) == 1
        ]
        for t in single_entry_transitions:
            source_place = list(t.in_arcs)[0].source
            # We check:
            # 1) source_place has exactly 1 input transition
            # 2) source_place is used *only* by 't' (i.e., no other transitions).
            if (
                len(source_place.in_arcs) == 1
                and _has_only_this_transition_as_output(source_place, t)
            ):
                # We can merge out 't' and 'source_place'
                source_transition = list(source_place.in_arcs)[0].source
                target_places = [arc.target for arc in t.out_arcs]

                remove_transition(net, t)
                remove_place(net, source_place)

                # Wire the old source transition directly to t's old targets
                for p in target_places:
                    add_arc_from_to(source_transition, p, net)

                cont = True
                break
    return net


def reduce_single_exit_transitions(net):
    """
    Reduces the number of single-exit silent transitions in the Petri net
    by merging them downward when the place receiving them does not receive
    arcs from any other transition.

    Parameters
    --------------
    net
        Petri net
    """
    cont = True
    while cont:
        cont = False
        # Identify silent transitions with exactly 1 output place.
        single_exit_transitions = [
            t for t in net.transitions
            if t.label is None and len(t.out_arcs) == 1
        ]
        for t in single_exit_transitions:
            target_place = list(t.out_arcs)[0].target
            # We check:
            # 1) target_place has exactly 1 output transition
            # 2) target_place is used *only* by 't' (i.e., no other transitions).
            if (
                len(target_place.out_arcs) == 1
                and _has_only_this_transition_as_input(target_place, t)
            ):
                target_transition = list(target_place.out_arcs)[0].target
                source_places = [arc.source for arc in t.in_arcs]

                remove_transition(net, t)
                remove_place(net, target_place)

                # Wire the old input places directly to t's old target transition
                for p in source_places:
                    add_arc_from_to(p, target_transition, net)

                cont = True
                break
    return net


def apply_simple_reduction(net):
    """
    Apply a simple reduction to the Petri net by repeatedly merging
    single-entry and single-exit silent transitions where possible.

    Parameters
    --------------
    net
        Petri net
    """
    changed = True
    while changed:
        old_t_count = len(net.transitions)
        old_p_count = len(net.places)

        reduce_single_entry_transitions(net)
        reduce_single_exit_transitions(net)

        new_t_count = len(net.transitions)
        new_p_count = len(net.places)

        # If the overall structure changed, keep iterating
        changed = (new_t_count < old_t_count or new_p_count < old_p_count)

    return net


######################################################################
# REMAINING REDUCTION FUNCTIONS (unchanged)
######################################################################

def apply_fst_rule(net):
    """
    Apply the Fusion of Series Transitions (FST) rule

    Parameters
    --------------
    net
        Reset Inhibitor net
    """
    cont = True
    while cont:
        cont = False
        for p, t, u in itertools.product(
            net.places, net.transitions, net.transitions
        ):
            if (
                (len(list(p.in_arcs)) == 1 and list(p.in_arcs)[0].source == t)
                and (
                    len(list(p.out_arcs)) == 1
                    and list(p.out_arcs)[0].target == u
                )
                and (
                    len(list(u.in_arcs)) == 1
                    and list(u.in_arcs)[0].source == p
                )
                and (len(post_set(t).intersection(post_set(u))) == 0)
                and (
                    set().union(
                        *[
                            post_set(place, properties.INHIBITOR_ARC)
                            for place in post_set(u)
                        ]
                    )
                    == post_set(p, properties.INHIBITOR_ARC)
                )
                and (
                    set().union(
                        *[
                            post_set(place, properties.RESET_ARC)
                            for place in post_set(u)
                        ]
                    )
                    == post_set(p, properties.RESET_ARC)
                )
                and (
                    len(pre_set(u, properties.RESET_ARC)) == 0
                    and len(pre_set(u, properties.INHIBITOR_ARC)) == 0
                )
            ):
                if u.label is None:
                    remove_place(net, p)
                    for target in post_set(u):
                        add_arc_from_to(t, target, net)
                    remove_transition(net, u)
                    cont = True
                    break

    return net


def apply_fsp_rule(net, im=None, fm=None):
    """
    Apply the Fusion of Series Places (FSP) rule

    Parameters
    --------------
    net
        Reset Inhibitor net
    """
    if im is None:
        im = {}
    if fm is None:
        fm = {}
    cont = True
    while cont:
        cont = False
        for p, q, t in itertools.product(
            net.places, net.places, net.transitions
        ):
            if (
                t.label is None
            ):  # only silent transitions may be removed either way
                if (
                    (len(t.in_arcs) == 1 and list(t.in_arcs)[0].source == p)
                    and (
                        len(t.out_arcs) == 1
                        and list(t.out_arcs)[0].target == q
                    )
                    and (len(post_set(p)) == 1 and list(post_set(p))[0] == t)
                    and (len(pre_set(p).intersection(pre_set(q))) == 0)
                    and (
                        post_set(p, properties.RESET_ARC)
                        == post_set(q, properties.RESET_ARC)
                    )
                    and (
                        post_set(p, properties.INHIBITOR_ARC)
                        == post_set(q, properties.INHIBITOR_ARC)
                    )
                    and (
                        len(pre_set(t, properties.RESET_ARC)) == 0
                        and len(pre_set(t, properties.INHIBITOR_ARC)) == 0
                    )
                ):
                    # remove place p and transition t
                    remove_transition(net, t)
                    for source in pre_set(p):
                        add_arc_from_to(source, q, net)
                    remove_place(net, p)
                    if p in im:
                        del im[p]
                        im[q] = 1
                    cont = True
                    break

    return net, im, fm


def apply_fpt_rule(net):
    """
    Apply the Fusion of Parallel Transitions (FPT) rule

    Parameters
    --------------
    net
        Reset Inhibitor net
    """
    cont = True
    while cont:
        cont = False
        # only silent transitions are considered for parallel merging
        silent_transitions = [
            transition for transition in net.transitions
            if transition.label is None
        ]
        for V in power_set(silent_transitions, 2):
            condition = True
            for x, y in itertools.product(V, V):
                if x != y:
                    if not (
                        (pre_set(x) == pre_set(y))
                        and (post_set(x) == post_set(y))
                        and (
                            pre_set(x, properties.RESET_ARC)
                            == pre_set(y, properties.RESET_ARC)
                        )
                        and (
                            pre_set(x, properties.INHIBITOR_ARC)
                            == pre_set(y, properties.INHIBITOR_ARC)
                        )
                    ):
                        condition = False
                        break
            # V is a valid candidate
            if condition and len(V) > 1:
                # remove transitions except the first one
                for t in V[1:]:
                    remove_transition(net, t)
                cont = True
                break
    return net


def apply_fpp_rule(net, im=None):
    """
    Apply the Fusion of Parallel Places (FPP) rule

    Parameters
    --------------
    net
        Reset Inhibitor net
    """
    if im is None:
        im = {}
    cont = True
    while cont:
        cont = False
        for Q in power_set(net.places, 2):
            condition = True
            for x, y in itertools.product(Q, Q):
                if x != y:
                    if not (
                        (pre_set(x) == pre_set(y))
                        and (post_set(x) == post_set(y))
                        and (
                            post_set(x, properties.RESET_ARC)
                            == post_set(y, properties.RESET_ARC)
                        )
                        and (
                            post_set(x, properties.INHIBITOR_ARC)
                            == post_set(y, properties.INHIBITOR_ARC)
                        )
                    ):
                        condition = False
                        break

            for x in Q:
                if x in im:
                    for y in Q:
                        if y in im and im[x] > im[y]:
                            if not (
                                len(post_set(x, properties.INHIBITOR_ARC)) == 0
                            ):
                                condition = False
                                break
                    else:
                        continue
                    break
            # Q is a valid candidate
            if condition and len(Q) > 1:
                # remove places except the first one
                for p in Q[1:]:
                    remove_place(net, p)
                cont = True
                break
    return net


def apply_elt_rule(net):
    """
    Apply the Elimination of Self-Loop Transitions (ELT) rule

    Parameters
    --------------
    net
        Reset Inhibitor net
    """
    cont = True
    while cont:
        cont = False
        for p, t in itertools.product(
            net.places, [t for t in net.transitions if t.label is None]
        ):
            if (
                (len(list(t.in_arcs)) == 1 and list(t.in_arcs)[0].source == p)
                and (
                    len(list(t.out_arcs)) == 1
                    and list(t.out_arcs)[0].target == p
                )
                and (len(list(p.in_arcs)) >= 2)
                and (
                    len(pre_set(t, properties.RESET_ARC)) == 0
                    and len(pre_set(t, properties.INHIBITOR_ARC)) == 0
                )
            ):
                remove_transition(net, t)
                cont = True
                break

    return net


def apply_elp_rule(net, im=None):
    """
    Apply the Elimination of Self-Loop Places (ELP) rule

    Parameters
    --------------
    net
        Reset Inhibitor net
    """
    if im is None:
        im = {}
    cont = True
    while cont:
        cont = False
        for p in [place for place in net.places]:
            if (
                (
                    set([arc.target for arc in p.out_arcs]).issubset(
                        set([arc.source for arc in p.in_arcs])
                    )
                )
                and (p in im and im[p] >= 1)
                and (
                    post_set(p, properties.RESET_ARC).union(
                        set([arc.target for arc in p.out_arcs])
                    )
                    == set([arc.source for arc in p.in_arcs])
                )
                and (len(post_set(p, properties.INHIBITOR_ARC)) == 0)
            ):
                remove_place(net, p)
                cont = True
                break

    return net


def apply_a_rule(net):
    """
    Apply the Abstraction (A) rule

    Parameters
    --------------
    net
        Reset Inhibitor net
    """
    cont = True
    while cont:
        cont = False
        for Q, U in itertools.product(
            power_set(net.places, 1), power_set(net.transitions, 1)
        ):
            for s, t in itertools.product(
                [s for s in net.places if s not in Q],
                [
                    t
                    for t in net.transitions
                    if (t not in U) and (t.label is None)
                ],
            ):
                if (
                    (
                        (pre_set(t) == {s})
                        and (post_set(s) == {t})
                        and (pre_set(s) == set(U))
                        and (post_set(t) == set(Q))
                        and (
                            len(
                                set(
                                    itertools.product(pre_set(s), post_set(t))
                                ).intersection(
                                    set(
                                        [
                                            (arc.source, arc.target)
                                            for arc in net.arcs
                                            if get_arc_type(arc) is None
                                        ]
                                    )
                                )
                            )
                        )
                        == 0
                    )
                    and (len(pre_set(t, properties.RESET_ARC)) == 0)
                    and (len(pre_set(t, properties.INHIBITOR_ARC)) == 0)
                ):

                    # check conditions on Q
                    condition = True
                    for q in Q:
                        if not (
                            (
                                post_set(s, properties.RESET_ARC)
                                == post_set(q, properties.RESET_ARC)
                            )
                            and (
                                post_set(s, properties.INHIBITOR_ARC)
                                == post_set(q, properties.INHIBITOR_ARC)
                            )
                        ):
                            condition = False
                            break
                    # Q is a valid candidate
                    if condition:
                        for u in U:
                            for q in Q:
                                add_arc_from_to(u, q, net)
                        remove_place(net, s)
                        remove_transition(net, t)
                        cont = True
                        break
    return net


def apply_r_rule(net):
    """
    Apply the Reset Reduction (R) rule

    Parameters
    --------------
    net
        Reset Inhibitor net
    """
    cont = True
    while cont:
        cont = False
        for p, t in itertools.product(net.places, net.transitions):
            if p in pre_set(t, properties.RESET_ARC).intersection(
                pre_set(t, properties.INHIBITOR_ARC)
            ):
                for arc in [arc for arc in net.arcs]:
                    if (
                        arc.source == p
                        and arc.target == t
                        and arc.arc_type == properties.RESET_ARC
                    ):
                        remove_arc(net, arc)
                cont = True
                break

    return net


def power_set(iterable, min=0):
    s = list(iterable)
    return chain.from_iterable(
        combinations(s, r) for r in range(min, len(s) + 1)
    )


def apply_reset_inhibitor_net_reduction(net, im=None, fm=None):
    """
    Apply a thorough reduction to the Reset Inhibitor net

    Parameters
    --------------
    net
        Reset Inhibitor net
    """
    if im is None:
        im = {}
    if fm is None:
        fm = {}
    apply_fst_rule(net)
    apply_fsp_rule(net, im, fm)
    # (FPT) and (FPP) and (A) rules may be very expensive for large nets
    # Uncomment them if needed:
    # apply_fpt_rule(net)
    # apply_fpp_rule(net, im)
    # apply_a_rule(net)
    apply_elt_rule(net)
    apply_elp_rule(net, im)
    apply_r_rule(net)
    return net, im, fm



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
class Trie(object):

    def __init__(
        self, label=None, parent=None, children=None, final=False, depth=0
    ):
        self._label = label
        self._parent = parent
        self._children = children if children is not None else list()
        self._final = final
        self._depth = depth

    def _set_parent(self, parent):
        self._parent = parent

    def _set_label(self, label):
        self._label = label

    def _set_children(self, children):
        self._children = children

    def _set_final(self, final):
        self._final = final

    def _get_children(self):
        return self._children

    def _get_final(self):
        return self._final

    def _get_parent(self):
        return self._parent

    def _get_label(self):
        return self._label

    def _set_depth(self, depth):
        self._depth = depth

    def _get_depth(self):
        return self._depth

    parent = property(_get_parent, _set_parent)
    children = property(_get_children, _set_children)
    label = property(_get_label, _set_label)
    final = property(_get_final, _set_final)
    depth = property(_get_depth, _set_depth)

    def repr_trie(self, indent_level=0):
        stri = []

        if self.label:
            stri.append("\t" * indent_level + self.label)
            indent_level += 1
        for child in self.children:
            stri.append(child.repr_trie(indent_level=indent_level))
        if self.final:
            stri.append("\t" * indent_level + "-- END --")

        return "\n".join(stri)

    def __repr__(self):
        return self.repr_trie()

    def __str__(self):
        return self.repr_trie()

    def __eq__(self, other):
        if not isinstance(other, Trie):
            return False
        if self._label != other._label:
            return False
        if self._final != other._final:
            return False
        this_children = sorted(self._children, key=lambda x: x._label)
        other_children = sorted(other._children, key=lambda x: x._label)
        if len(this_children) != len(other_children):
            return False
        for i in range(len(this_children)):
            if this_children[i] != other_children[i]:
                return False
        return True

    def __hash__(self):
        return hash(
            (
                self._label,
                self._final,
                tuple(sorted(self._children, key=lambda x: x._label)),
            )
        )



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or 
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from pm4py.util.xes_constants import DEFAULT_TIMESTAMP_KEY
import statistics
from pm4py.util.business_hours import BusinessHours
from pm4py.util import exec_utils, constants
from enum import Enum
from typing import Optional, Dict, Any, Union
from pm4py.objects.log.obj import EventLog
from pm4py.objects.conversion.log import converter as log_converter


class Parameters(Enum):
    ATTRIBUTE_KEY = constants.PARAMETER_CONSTANT_ATTRIBUTE_KEY
    ACTIVITY_KEY = constants.PARAMETER_CONSTANT_ACTIVITY_KEY
    TIMESTAMP_KEY = constants.PARAMETER_CONSTANT_TIMESTAMP_KEY
    CASE_ID_KEY = constants.PARAMETER_CONSTANT_CASEID_KEY

    BUSINESS_HOURS = "business_hours"
    BUSINESS_HOUR_SLOTS = "business_hour_slots"
    WORKCALENDAR = "workcalendar"


def get_case_arrival_avg(
    log: EventLog,
    parameters: Optional[Dict[Union[str, Parameters], Any]] = None,
) -> float:
    """
    Gets the average time interlapsed between case starts

    Parameters
    --------------
    log
        Trace log
    parameters
        Parameters of the algorithm, including:
            Parameters.TIMESTAMP_KEY -> attribute of the log to be used as timestamp

    Returns
    --------------
    case_arrival_avg
        Average time interlapsed between case starts
    """
    if parameters is None:
        parameters = {}

    log = log_converter.apply(
        log, variant=log_converter.Variants.TO_EVENT_LOG, parameters=parameters
    )

    business_hours = exec_utils.get_param_value(
        Parameters.BUSINESS_HOURS, parameters, False
    )
    business_hours_slots = exec_utils.get_param_value(
        Parameters.BUSINESS_HOUR_SLOTS,
        parameters,
        constants.DEFAULT_BUSINESS_HOUR_SLOTS,
    )
    workcalendar = exec_utils.get_param_value(
        Parameters.WORKCALENDAR,
        parameters,
        constants.DEFAULT_BUSINESS_HOURS_WORKCALENDAR,
    )

    timestamp_key = exec_utils.get_param_value(
        Parameters.TIMESTAMP_KEY, parameters, DEFAULT_TIMESTAMP_KEY
    )

    case_start_time = [
        trace[0][timestamp_key]
        for trace in log
        if trace and timestamp_key in trace[0]
    ]
    case_start_time = sorted(case_start_time)

    case_diff_start_time = []
    for i in range(len(case_start_time) - 1):
        if business_hours:
            bh = BusinessHours(
                case_start_time[i],
                case_start_time[i + 1],
                business_hour_slots=business_hours_slots,
                workcalendar=workcalendar,
            )
            case_diff_start_time.append(bh.get_seconds())
        else:
            case_diff_start_time.append(
                (case_start_time[i + 1] - case_start_time[i]).total_seconds()
            )

    if case_diff_start_time:
        return statistics.mean(case_diff_start_time)

    return 0.0


def get_case_dispersion_avg(
    log: EventLog,
    parameters: Optional[Dict[Union[str, Parameters], Any]] = None,
) -> float:
    """
    Gets the average time interlapsed between case ends

    Parameters
    --------------
    log
        Trace log
    parameters
        Parameters of the algorithm, including:
            Parameters.TIMESTAMP_KEY -> attribute of the log to be used as timestamp

    Returns
    --------------
    case_dispersion_avg
        Average time interlapsed between the completion of cases
    """
    if parameters is None:
        parameters = {}

    log = log_converter.apply(
        log, variant=log_converter.Variants.TO_EVENT_LOG, parameters=parameters
    )

    business_hours = exec_utils.get_param_value(
        Parameters.BUSINESS_HOURS, parameters, False
    )
    business_hours_slots = exec_utils.get_param_value(
        Parameters.BUSINESS_HOUR_SLOTS,
        parameters,
        constants.DEFAULT_BUSINESS_HOUR_SLOTS,
    )

    workcalendar = exec_utils.get_param_value(
        Parameters.WORKCALENDAR,
        parameters,
        constants.DEFAULT_BUSINESS_HOURS_WORKCALENDAR,
    )

    timestamp_key = exec_utils.get_param_value(
        Parameters.TIMESTAMP_KEY, parameters, DEFAULT_TIMESTAMP_KEY
    )

    case_end_time = [
        trace[-1][timestamp_key]
        for trace in log
        if trace and timestamp_key in trace[0]
    ]
    case_end_time = sorted(case_end_time)

    case_diff_end_time = []
    for i in range(len(case_end_time) - 1):
        if business_hours:
            bh = BusinessHours(
                case_end_time[i],
                case_end_time[i + 1],
                business_hour_slots=business_hours_slots,
                workcalendar=workcalendar,
            )
            case_diff_end_time.append(bh.get_seconds())
        else:
            case_diff_end_time.append(
                (case_end_time[i + 1] - case_end_time[i]).total_seconds()
            )

    if case_diff_end_time:
        return statistics.mean(case_diff_end_time)

    return 0.0



'''
    PM4Py â€“ A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschrÃ¤nkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from pm4py.algo.evaluation.simplicity.variants import (
    arc_degree,
    extended_cardoso,
    extended_cyclomatic,
)
from enum import Enum
from pm4py.util import exec_utils
from typing import Optional, Dict, Any
from pm4py.objects.petri_net.obj import PetriNet


class Variants(Enum):
    SIMPLICITY_ARC_DEGREE = arc_degree
    EXTENDED_CARDOSO = extended_cardoso
    EXTENDED_CYCLOMATIC = extended_cyclomatic


SIMPLICITY_ARC_DEGREE = Variants.SIMPLICITY_ARC_DEGREE
EXTENDED_CARDOSO = Variants.EXTENDED_CARDOSO
EXTENDED_CYCLOMATIC = Variants.EXTENDED_CYCLOMATIC

VERSIONS = {SIMPLICITY_ARC_DEGREE, EXTENDED_CARDOSO, EXTENDED_CYCLOMATIC}


def apply(
    petri_net: PetriNet,
    parameters: Optional[Dict[Any, Any]] = None,
    variant=SIMPLICITY_ARC_DEGREE,
) -> float:
    return exec_utils.get_variant(variant).apply(
        petri_net, parameters=parameters
    )



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from enum import Enum
from pm4py.util import exec_utils
from pm4py.objects.ocel import constants as ocel_constants
from pm4py.objects.ocel.obj import OCEL
from typing import Optional, Dict, Any
from copy import copy
from pm4py.objects.ocel.util import filtering_utils


class Parameters(Enum):
    EVENT_ID = ocel_constants.PARAM_EVENT_ID
    OBJECT_ID = ocel_constants.PARAM_OBJECT_ID
    OBJECT_TYPE = ocel_constants.PARAM_OBJECT_TYPE


def filter_start_events_per_object_type(
    ocel: OCEL, object_type: str, parameters: Optional[Dict[Any, Any]] = None
) -> OCEL:
    """
    Filters the events in which a new object for the given object type is spawn.
    (E.g. an event with activity "Create Order" might spawn new orders).

    Parameters
    ------------------
    ocel
        Object-centric event log
    object_type
        Object type to consider
    parameters
        Parameters of the algorithm, including:
        - Parameters.EVENT_ID => the attribute working as event identifier
        - Parameters.OBJECT_ID => the attribute working as object identifier
        - Parameters.OBJECT_TYPE => the attribute working as object type

    Returns
    ------------------
    filtered_ocel
        Filtered object-centric event log
    """
    if parameters is None:
        parameters = {}

    event_id = exec_utils.get_param_value(
        Parameters.EVENT_ID, parameters, ocel.event_id_column
    )
    object_id = exec_utils.get_param_value(
        Parameters.OBJECT_ID, parameters, ocel.object_id_column
    )
    object_type_column = exec_utils.get_param_value(
        Parameters.OBJECT_TYPE, parameters, ocel.object_type_column
    )

    evs = (
        ocel.relations[ocel.relations[object_type_column] == object_type]
        .groupby(object_id)
        .first()[event_id]
        .to_numpy()
        .tolist()
    )

    ocel = copy(ocel)
    ocel.events = ocel.events[ocel.events[event_id].isin(evs)]

    return filtering_utils.propagate_event_filtering(
        ocel, parameters=parameters
    )


def filter_end_events_per_object_type(
    ocel: OCEL, object_type: str, parameters: Optional[Dict[Any, Any]] = None
) -> OCEL:
    """
    Filters the events in which an object for the given object type terminates its lifecycle.
    (E.g. an event with activity "Pay Order" might terminate an order).

    Parameters
    ------------------
    ocel
        Object-centric event log
    object_type
        Object type to consider
    parameters
        Parameters of the algorithm, including:
        - Parameters.EVENT_ID => the attribute working as event identifier
        - Parameters.OBJECT_ID => the attribute working as object identifier
        - Parameters.OBJECT_TYPE => the attribute working as object type

    Returns
    ------------------
    filtered_ocel
        Filtered object-centric event log
    """
    if parameters is None:
        parameters = {}

    event_id = exec_utils.get_param_value(
        Parameters.EVENT_ID, parameters, ocel.event_id_column
    )
    object_id = exec_utils.get_param_value(
        Parameters.OBJECT_ID, parameters, ocel.object_id_column
    )
    object_type_column = exec_utils.get_param_value(
        Parameters.OBJECT_TYPE, parameters, ocel.object_type_column
    )

    evs = (
        ocel.relations[ocel.relations[object_type_column] == object_type]
        .groupby(object_id)
        .last()[event_id]
    )

    ocel = copy(ocel)
    ocel.events = ocel.events[ocel.events[event_id].isin(evs)]

    return filtering_utils.propagate_event_filtering(
        ocel, parameters=parameters
    )



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or 
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''

from pm4py.util.lp import solver
from pm4py.util import constants
from pm4py.objects.petri_net.utils.petri_utils import remove_place
from copy import copy
from typing import Tuple
import warnings
import importlib.util
from pm4py.objects.petri_net.obj import PetriNet, Marking


def apply_reduction(
    net: PetriNet, im: Marking, fm: Marking
) -> Tuple[PetriNet, Marking, Marking]:
    """
    Apply the Murata reduction to an accepting Petri net, removing the structurally redundant places.

    The implementation follows the Berthelot algorithm as in:
    https://svn.win.tue.nl/repos/prom/Packages/Murata/Trunk/src/org/processmining/algorithms/BerthelotAlgorithm.java

    Parameters
    ---------------
    net
        Petri net
    im
        Initial marking
    fm
        Final marking

    Returns
    --------------
    net
        Petri net
    im
        Initial marking
    fm
        Final marking
    """
    places = sorted(list(net.places), key=lambda x: x.name)
    redundant = set()
    for place in places:
        # Skip places in the initial or final markings
        if place in im or place in fm:
            continue

        Aeq = []
        Aub = []
        beq = []
        bub = []

        # first constraint
        constraint = [0] * (len(net.places) + 1)
        for p2 in im:
            if p2 not in redundant:
                if p2 == place:
                    constraint[places.index(p2)] = im[p2]
                else:
                    constraint[places.index(p2)] = -im[p2]
        constraint[-1] = -1
        Aeq.append(constraint)
        beq.append(0)

        # second constraints
        for trans in net.transitions:
            constraint = [0] * (len(net.places) + 1)

            for arc in trans.in_arcs:
                p2 = arc.source
                if p2 not in redundant:
                    if p2 == place:
                        constraint[places.index(p2)] = arc.weight
                    else:
                        constraint[places.index(p2)] = -arc.weight
            constraint[-1] = -1
            Aub.append(constraint)
            bub.append(0)

        # third constraints
        for trans in net.transitions:
            constraint = [0] * (len(net.places) + 1)

            for arc in trans.out_arcs:
                p2 = arc.target
                if p2 not in redundant:
                    if p2 == place:
                        constraint[places.index(p2)] = -arc.weight
                    else:
                        constraint[places.index(p2)] = arc.weight
            Aub.append(constraint)
            bub.append(0)

        # fourth constraint
        for p2 in net.places:
            if p2 not in redundant:
                constraint = [0] * (len(net.places) + 1)

                constraint[places.index(p2)] = -1

                Aub.append(constraint)
                if p2 == place:
                    bub.append(-1)
                else:
                    bub.append(0)

        # fifth constraint
        constraint = [0] * (len(net.places) + 1)
        constraint[-1] = -1
        Aub.append(constraint)
        bub.append(0)

        c = [1] * (len(net.places) + 1)
        integrality = [1] * (len(net.places) + 1)

        proposed_solver = solver.SCIPY
        if importlib.util.find_spec("pulp"):
            proposed_solver = solver.PULP
        else:
            if constants.SHOW_INTERNAL_WARNINGS:
                warnings.warn(
                    "solution from scipy may be unstable. Please install PuLP (pip install pulp) for fully reliable results."
                )

        xx = solver.apply(
            c,
            Aub,
            bub,
            Aeq,
            beq,
            variant=proposed_solver,
            parameters={"integrality": integrality},
        )

        if (hasattr(xx, "success") and xx.success) or (
            hasattr(xx, "sol_status") and xx.sol_status > -1
        ):
            redundant.add(place)

    for place in redundant:
        # Remove the redundant place from the net
        net = remove_place(net, place)

    return net, im, fm
